import unittest
import os
from deepchem.utils.data_utils import load_sdf_files


class TestFileLoading(unittest.TestCase):

    def test_load_sdf_files(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        file_path = [os.path.join(current_dir, 'assets', 'gdb9_small.sdf')]
        for df in load_sdf_files(file_path):
            break
        df_shape = (2, 6)
        self.assertEqual(df.shape, df_shape)
        self.assertEqual(df['smiles'][0], '[H]C([H])([H])[H]')
        n_atoms_mol1 = 5
        self.assertEqual(df['mol'][0].GetNumAtoms(), n_atoms_mol1)
        self.assertEqual(len(eval(df['pos_x'][0])), n_atoms_mol1)
        self.assertEqual(len(eval(df['pos_y'][0])), n_atoms_mol1)
        self.assertEqual(len(eval(df['pos_y'][0])), n_atoms_mol1)



import numpy as np
from collections import defaultdict
from deepchem.feat import Featurizer
from typing import List, Dict, Tuple, Any


class _Realigner(object):

    def left_align_indel(self, seq: str, pos: int,
                         indel: str) -> Tuple[int, str]:
        """
        Left align an indel by shifting it to the left as
        much as possible.

        Parameters
        ----------

        seq : str
            Reference sequence as a string.
        pos : int
            Position of the indel.
        indel : str
            Indel sequence (e.g., "+2AT" or "-2").

        Returns
        -------

        Tuple[int, str]
            New position and indel string.

        """
        if indel.startswith('+'):
            indel_seq = indel[2:]
            while pos > 0 and seq[pos - 1] == indel_seq[-1]:
                pos -= 1
                indel_seq = seq[pos] + indel_seq[:-1]
            return pos, f"+{len(indel_seq)}{indel_seq}"
        elif indel.startswith('-'):
            del_len = int(indel[1:])
            while pos > 0 and seq[pos - 1] == seq[pos + del_len - 1]:
                pos -= 1
            return pos, f"-{del_len}"
        return pos, indel

    def decode_one_hot(self,
                       one_hot_vector: List[np.ndarray],
                       charset: List[str] = ["A", "C", "T", "G", "N"]) -> str:
        """
        Decode a one-hot encoded sequence into a string of
        nucleotides.

        Parameters
        ----------

        one_hot_vector : List[np.ndarray]
            List of one-hot encoded vectors.
        charset : Optional[List[str]]
            List of characters corresponding to the encoding.
            Default is ["A", "C", "T", "G", "N"].

        Returns
        -------

        str
            Decoded sequence as a string.

        """
        decoded_seq = []
        for vector in one_hot_vector:
            idx = np.argmax(vector)
            decoded_seq.append(charset[idx])
        return ''.join(decoded_seq)

    def process_pileup(
            self, pileup_info: List[Dict[str,
                                         Any]], reference_seq_dict: Dict[str,
                                                                         str],
            allele_counts: Dict[Tuple[str, int], Dict[str, Any]]) -> None:
        """
        Process pileup information to extract allele counts.

        Parameters
        ----------

        pileup_info : List[Dict[str, Any]]
            List of dictionaries containing pileup information.
        reference_seq_dict : Dict[str, str]
            Dictionary with chromosome names as keys and reference
            sequences as values.
        allele_counts : Dict[Tuple[str, int], Dict[str, Any]]
            Dictionary to store allele counts.

        """
        for pileupcolumn in pileup_info:
            ref_base = reference_seq_dict[pileupcolumn['name']][
                pileupcolumn['pos']]
            allele_count: Dict[str, Any] = {
                "reference_base": ref_base,
                "read_alleles": defaultdict(int),
                "coverage": 0
            }
            for read_seq in pileupcolumn['reads']:
                if read_seq[2]:
                    base = '-'
                    print(allele_count)
                    allele_count["read_alleles"][base] += 1
                    allele_count["coverage"] += 1
                elif read_seq[3]:
                    base = 'N'
                    allele_count["read_alleles"][base] += 1
                    allele_count["coverage"] += 1
                else:
                    indel = ""
                    if read_seq[4] > 0:
                        indel_seq = read_seq[0][read_seq[2]:read_seq[2] +
                                                read_seq[4]]
                        indel = f"+{read_seq[4]}{indel_seq}"
                        new_pos, indel = self.left_align_indel(
                            reference_seq_dict[pileupcolumn['name']]
                            [0:pileupcolumn['pos'] + read_seq[4] + 1],
                            pileupcolumn['pos'], indel)
                    elif read_seq[4] < 0:
                        indel = f"{read_seq[4]}"
                        new_pos, indel = self.left_align_indel(
                            reference_seq_dict[pileupcolumn['name']]
                            [0:pileupcolumn['pos'] + read_seq[4] + 1],
                            pileupcolumn['pos'], indel)

                    base = read_seq[0][read_seq[1]]

                    if base in {'A', 'C', 'G', 'T'}:
                        if indel:
                            allele_count["read_alleles"][base + indel] += 1
                            allele_count["coverage"] += 1
                        else:
                            allele_count["read_alleles"][base] += 1
                            allele_count["coverage"] += 1
            pos = pileupcolumn['pos'] + 1
            allele_counts[(pileupcolumn['name'], pos)] = allele_count

    def generate_pileup_and_reads(
        self, bamfile_path: str, reference_path: str
    ) -> Tuple[Dict[Tuple[str, int], Dict[str, Any]], List[Any]]:
        """
        Generate pileup and reads from BAM and reference FASTA files.

        Parameters
        ----------

        bamfile_path : str
            Path to the BAM file.
        reference_path : str
            Path to the reference FASTA file.

        Returns
        -------

        Tuple[Dict[Tuple[str, int], Dict[str, Any]], List[Any]]
            Dictionary of allele counts and list of reads.

        """
        from deepchem.data import FASTALoader, BAMLoader

        bam_loader = BAMLoader(get_pileup=True)
        bam_dataset = bam_loader.create_dataset(bamfile_path)

        fasta_loader = FASTALoader(None, False, False)
        fasta_dataset = fasta_loader.create_dataset(reference_path)

        allele_counts: Dict[Tuple[str, int], Dict[str, Any]] = {}
        reads = []

        one_hot_encoded_sequences = fasta_dataset.X
        decoded_sequences: List[str] = []

        # Convert the one-hot encoded sequences to strings
        for seq in one_hot_encoded_sequences:
            decoded_seq = self.decode_one_hot(seq)
            decoded_sequences.append(decoded_seq)

        # Map the sequences to chrom names
        chrom_names = ["chr1", "chr2"]

        reference_seq_dict: Dict[str, str] = {
            chrom_names[i]: seq for i, seq in enumerate(decoded_sequences)
        }

        for x, y, w, ids in bam_dataset.itersamples():
            chrom = x[3]  # Reference name

            pileup_info = x[7] if len(x) > 7 else None
            for pileupcolumn in pileup_info:
                for read_seq in pileupcolumn['reads']:
                    reads.append(read_seq)

            if chrom not in reference_seq_dict:
                continue

            if pileup_info:
                self.process_pileup(pileup_info, reference_seq_dict,
                                    allele_counts)

        return allele_counts, reads

    def update_counts(self, count: int, start: int, end: int,
                      window_counts: Dict[int, int]) -> None:
        """
        Update counts in a window.

        Parameters
        ----------

        count : int
            Count to add.
        start : int
            Start position of the window.
        end : int
            End position of the window.
        window_counts : Dict[int, int]
            Dictionary to store window counts.

        """
        for pos in range(start, end):
            window_counts[pos] += count

    def select_candidate_regions(
        self, allele_counts: Dict[Tuple[str, int], Dict[str, Any]]
    ) -> List[Tuple[str, int, int, int]]:
        """
        Select candidate regions based on allele counts.

        Parameters
        ----------

        allele_counts : Dict[Tuple[str, int], Dict[str, Any]]
            Dictionary of allele counts.

        Returns
        -------

        List[Tuple[str, int, int, int]]
            List of candidate regions.

        """
        window_counts: Dict[int, int] = defaultdict(int)
        chrom_positions = defaultdict(list)

        for (chrom, pos), allele_count in allele_counts.items():
            for allele, count in allele_count["read_alleles"].items():
                if allele != allele_count["reference_base"]:
                    allele_type = ""
                    if allele.startswith("+") or allele.startswith("-"):
                        allele_length = int(allele[1:])
                        allele_type = "INSERTION" if allele.startswith(
                            "+") else "DELETION"
                    elif allele == "N":
                        allele_type = "SOFT_CLIP"
                        allele_length = 1
                    else:
                        allele_type = "SUBSTITUTION"
                        allele_length = 1

                    start, end = 0, 0

                    if allele_type == "SUBSTITUTION":
                        start = pos
                        end = pos + 1
                    elif allele_type in ["SOFT_CLIP", "INSERTION"]:
                        start = pos + 1 - (allele_length - 1)
                        end = pos + allele_length
                    elif allele_type == "DELETION":
                        start = pos + 1
                        end = pos + allele_length

                    self.update_counts(count, start, end, window_counts)
                    chrom_positions[chrom].append((start, end))

        candidate_regions = []
        for chrom, positions in chrom_positions.items():
            positions.sort()
            if positions:
                start, end = positions[0]
                current_count = sum(
                    window_counts[pos] for pos in range(start, end))

                for (next_start, next_end) in positions[1:]:
                    if next_start <= end + 1:
                        end = max(end, next_end)
                        current_count += sum(
                            window_counts[pos]
                            for pos in range(next_start, next_end))
                    else:
                        candidate_regions.append(
                            (chrom, start, end, current_count))
                        start, end = next_start, next_end
                        current_count = sum(
                            window_counts[pos] for pos in range(start, end))

                candidate_regions.append((chrom, start, end, current_count))

        return candidate_regions


class RealignerFeaturizer(Featurizer):
    """
    Realigns reads and generates candidate regions for variant calling.

    More features can be added to this class in the future.

    """

    def __init__(self):
        self.realigner = _Realigner()

    def _featurize(self, datapoint):
        """
        Featurizes a datapoint by generating candidate regions and reads.

        Args:
            datapoint (Tuple[str, str]): A tuple containing two strings
            representing allele counts and reads.

        Returns:
            Tuple[List[Tuple[str, int, int, int]], List[Any]]: A tuple
            containing the candidate regions and reads.

        """
        allele_counts, reads = self.realigner.generate_pileup_and_reads(
            datapoint[0], datapoint[1])
        candidate_regions = self.realigner.select_candidate_regions(
            allele_counts)
        return candidate_regions, reads



"""
Utilities for miscellaneous tasks.
"""
import functools
from typing import Dict, List, Mapping, Optional, Callable, TypeVar, Any


def indent(s, nspace):
    """Gives indentation of the second line and next lines.
    It is used to format the string representation of an object.
    Which might be containing multiples objects in it.
    Usage: LinearOperator

    Parameters
    ----------
    s: str
        The string to be indented.
    nspace: int
        The number of spaces to be indented.

    Returns
    -------
    str
        The indented string.

    """
    spaces = " " * nspace
    lines = [spaces + c if i > 0 else c for i, c in enumerate(s.split("\n"))]
    return "\n".join(lines)


def shape2str(shape):
    """Convert the shape to string representation.
    It also nicely formats the shape to be readable.

    Parameters
    ----------
    shape: Sequence[int]
        The shape to be converted to string representation.

    Returns
    -------
    str
        The string representation of the shape.

    """
    return "(%s)" % (", ".join([str(s) for s in shape]))


# Warnings
class UnimplementedError(Exception):
    """
    Raised if a method is not implemented.
    """
    pass


class GetSetParamsError(Exception):
    """
    Raised if there is an error in getting or setting parameters.
    """
    pass


class ConvergenceWarning(Warning):
    """
    Warning to be raised if the convergence of an algorithm is not achieved.
    """
    pass


class MathWarning(Warning):
    """
    Raised if there are mathematical conditions that are not satisfied.
    """
    pass


class Uniquifier(object):
    """
    Identifies and tracks unique objects within a list, even if they are
    duplicates based on internal memory addresses (using id()).
    It Optimizes operations involving unique objects by avoiding redundant
    processing.

    Examples
    --------
    >>> from deepchem.utils import Uniquifier
    >>> a = 1
    >>> b = 2
    >>> c = 3
    >>> d = 1
    >>> u = Uniquifier([a, b, c, a, d])
    >>> u.get_unique_objs()
    [1, 2, 3]

    """

    def __init__(self, allobjs: List):
        """Initialize the uniquifier.

        Parameters
        ----------
        allobjs: List
            The list of objects to be uniquified.

        """
        self.nobjs = len(allobjs)

        id2idx: Dict[int, int] = {}
        unique_objs: List[int] = []
        unique_idxs: List[int] = []
        nonunique_map_idxs: List[int] = [-self.nobjs * 2] * self.nobjs
        num_unique = 0
        for i, obj in enumerate(allobjs):
            id_obj = id(obj)
            if id_obj in id2idx:
                nonunique_map_idxs[i] = id2idx[id_obj]
                continue
            id2idx[id_obj] = num_unique
            unique_objs.append(obj)
            nonunique_map_idxs[i] = num_unique
            unique_idxs.append(i)
            num_unique += 1

        self.unique_objs = unique_objs
        self.unique_idxs = unique_idxs
        self.nonunique_map_idxs = nonunique_map_idxs
        self.num_unique = num_unique
        self.all_unique = self.nobjs == self.num_unique

    def get_unique_objs(self, allobjs: Optional[List] = None) -> List:
        """Get the unique objects.

        Parameters
        ----------
        allobjs: Optional[List]
            The list of objects to be uniquified.

        Returns
        -------
        List
            The list of unique objects.

        """
        if allobjs is None:
            return self.unique_objs
        assert len(
            allobjs
        ) == self.nobjs, "The allobjs must have %d elements" % self.nobjs
        if self.all_unique:
            return allobjs
        return [allobjs[i] for i in self.unique_idxs]

    def map_unique_objs(self, uniqueobjs: List) -> List:
        """Map the unique objects to the original objects.

        Parameters
        ----------
        uniqueobjs: List
            The list of unique objects.

        """
        assert len(
            uniqueobjs
        ) == self.num_unique, "The uniqueobjs must have %d elements" % self.num_unique
        if self.all_unique:
            return uniqueobjs
        return [uniqueobjs[idx] for idx in self.nonunique_map_idxs]


T = TypeVar('T')
K = TypeVar('K')


def memoize_method(fcn: Callable[[Any], T]) -> Callable[[Any], T]:
    """Memoize a method without any arguments using a cache in the object

    Examples
    --------
    >>> class A:
    ...     @memoize_method
    ...     def foo(self):
    ...         print("foo")
    ...         return 1
    >>> a = A()
    >>> a.foo()
    foo
    1
    >>> a.foo()
    1

    Parameters
    ----------
    fcn: callable
        Function to memoize

    Returns
    -------
    callable
        Memoized function

    """
    cachename = "__cch_" + fcn.__name__

    @functools.wraps(fcn)
    def new_fcn(self) -> T:
        if cachename in self.__dict__:
            return self.__dict__[cachename]
        else:
            res = fcn(self)
            self.__dict__[cachename] = res
            return res

    return new_fcn


def get_option(name: str, s: K, options: Mapping[K, T]) -> T:
    """Get the value from dictionary of options, if not found, then raise an error

    Examples
    --------
    >>> from deepchem.utils import get_option
    >>> options = {"a": 1, "b": 2}
    >>> get_option("name", "a", options)
    1

    Parameters
    ----------
    name : str
        Name of the option
    s : K
        Key to be searched
    options : Mapping[K, T]
        Dictionary of options

    Returns
    -------
    T
        Value of the option
    """
    if s in options:
        return options[s]
    else:
        raise ValueError(
            f"Unknown {name}: {s}. The available options are: {str(list(options.keys()))}"
        )



import warnings
import torch
from typing import Callable, List, Optional


def gd(
        fcn: Callable[..., torch.Tensor],
        x0: torch.Tensor,
        params: List,
        # gd parameters
        step: float = 1e-3,
        gamma: float = 0.9,
        # stopping conditions
        maxiter: int = 1000,
        f_tol: float = 0.0,
        f_rtol: float = 1e-8,
        x_tol: float = 0.0,
        x_rtol: float = 1e-8,
        # misc parameters
        verbose=False,
        **unused):
    r"""
    Vanilla gradient descent with momentum. The stopping conditions use OR criteria.
    The update step is following the equations below.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils.optimize.minimizer import gd
    >>> def fcn(x):
    ...     return (x - 2) ** 2, 2 * (x - 2)
    >>> x0 = torch.tensor(0.0, requires_grad=True)
    >>> x = gd(fcn, x0, [])
    >>> x
    tensor(2.0000)


    .. math::
        \mathbf{v}_{t+1} &= \gamma \mathbf{v}_t - \eta \nabla_{\mathbf{x}} f(\mathbf{x}_t) \\
        \mathbf{x}_{t+1} &= \mathbf{x}_t + \mathbf{v}_{t+1}

    Parameters
    ----------
    fcn: callable
        The objective function to minimize. It should take a tensor and return a
        tensor and its gradient.
    x0: torch.Tensor
        The initial guess.
    step: float
        The step size towards the steepest descent direction, i.e. :math:`\eta` in
        the equations above.
    gamma: float
        The momentum factor, i.e. :math:`\gamma` in the equations above.
    maxiter: int
        Maximum number of iterations.
    f_tol: float or None
        The absolute tolerance of the output ``f``.
    f_rtol: float or None
        The relative tolerance of the output ``f``.
    x_tol: float or None
        The absolute tolerance of the norm of the input ``x``.
    x_rtol: float or None
        The relative tolerance of the norm of the input ``x``.

    """

    x = x0.clone()
    stop_cond = TerminationCondition(f_tol, f_rtol, x_tol, x_rtol, verbose)
    fprev = torch.tensor(0.0, dtype=x0.dtype, device=x0.device)
    v = torch.zeros_like(x)
    for i in range(maxiter):
        f, dfdx = fcn(x, *params)

        # update the step
        v = (gamma * v - step * dfdx).detach()
        xprev = x.detach()
        x = (xprev + v).detach()

        # check the stopping conditions
        to_stop = stop_cond.to_stop(i, x, xprev, f, fprev)

        if to_stop:
            break

        fprev = f
    x = stop_cond.get_best_x(x)
    return x


def adam(
        fcn: Callable[..., torch.Tensor],
        x0: torch.Tensor,
        params: List,
        # gd parameters
        step: float = 1e-3,
        beta1: float = 0.9,
        beta2: float = 0.999,
        eps: float = 1e-8,
        # stopping conditions
        maxiter: int = 1000,
        f_tol: float = 0.0,
        f_rtol: float = 1e-8,
        x_tol: float = 0.0,
        x_rtol: float = 1e-8,
        # misc parameters
        verbose=False,
        **unused):
    r"""
    Adam optimizer by Kingma & Ba (2015). The stopping conditions use OR criteria.
    The update step is following the equations below.

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils.optimize.minimizer import adam
    >>> def fcn(x):
    ...     return (x - 4) * 2, (x * 2) + 3
    >>> x0 = torch.tensor(0.0, requires_grad=True)
    >>> x = adam(fcn, x0, [], maxiter=10000)
    >>> x
    tensor(-1.4999)

    .. math::
        \mathbf{g}_t &= \nabla_{\mathbf{x}} f(\mathbf{x}_{t-1}) \\
        \mathbf{m}_t &= \beta_1 \mathbf{m}_{t-1} + (1 - \beta_1) \mathbf{g}_t \\
        \mathbf{v}_t &= \beta_2 \mathbf{v}_{t-1} + (1 - \beta_2) \mathbf{g}_t^2 \\
        \hat{\mathbf{m}}_t &= \mathbf{m}_t / (1 - \beta_1^t) \\
        \hat{\mathbf{v}}_t &= \mathbf{v}_t / (1 - \beta_2^t) \\
        \mathbf{x}_t &= \mathbf{x}_{t-1} - \alpha \hat{\mathbf{m}}_t / (\sqrt{\hat{\mathbf{v}}_t} + \epsilon)

    Parameters
    ----------
    step: float
        The step size towards the descent direction, i.e. :math:`\alpha` in
        the equations above.
    beta1: float
        Exponential decay rate for the first moment estimate.
    beta2: float
        Exponential decay rate for the first moment estimate.
    eps: float
        Small number to prevent division by 0.
    maxiter: int
        Maximum number of iterations.
    f_tol: float or None
        The absolute tolerance of the output ``f``.
    f_rtol: float or None
        The relative tolerance of the output ``f``.
    x_tol: float or None
        The absolute tolerance of the norm of the input ``x``.
    x_rtol: float or None
        The relative tolerance of the norm of the input ``x``.
    """

    x = x0.clone()
    stop_cond = TerminationCondition(f_tol, f_rtol, x_tol, x_rtol, verbose)
    fprev = torch.tensor(0.0, dtype=x0.dtype, device=x0.device)
    v = torch.zeros_like(x)
    m = torch.zeros_like(x)
    beta1t = beta1
    beta2t = beta2
    for i in range(maxiter):
        f, dfdx = fcn(x, *params)
        f = f.detach()
        dfdx = dfdx.detach()

        # update the step
        m = beta1 * m + (1 - beta1) * dfdx
        v = beta2 * v + (1 - beta2) * dfdx**2
        mhat = m / (1 - beta1t)
        vhat = v / (1 - beta2t)
        beta1t *= beta1
        beta2t *= beta2
        xprev = x.detach()
        x = (xprev - step * mhat / (vhat**0.5 + eps)).detach()

        # check the stopping conditions
        to_stop = stop_cond.to_stop(i, x, xprev, f, fprev)

        if to_stop:
            break

        fprev = f
    x = stop_cond.get_best_x(x)
    return x


class TerminationCondition(object):
    """The class to handle the stopping conditions.

    Examples
    --------
    >>> stop_cond = TerminationCondition(1e-8, 1e-8, 1e-8, 1e-8, True)

    """

    def __init__(self, f_tol: float, f_rtol: float, x_tol: float, x_rtol: float,
                 verbose: bool):
        """Initialize the TerminationCondition.

        Parameters
        ----------
        f_tol: float or None
            Absolute tolerance of the output ``f``.
        f_rtol: float or None
            Relative tolerance of the output ``f``.
        x_tol: float or None
            Absolute tolerance of the norm of the input ``x``.
        x_rtol: float or None
            Relative tolerance of the norm of the input ``x``.
        verbose: bool
            Whether to print the iteration information.

        """
        self.f_tol = f_tol
        self.f_rtol = f_rtol
        self.x_tol = x_tol
        self.x_rtol = x_rtol
        self.verbose = verbose

        self._ever_converge = False
        self._max_i = -1
        self._best_dxnorm = float("inf")
        self._best_df = float("inf")
        self._best_f = float("inf")
        self._best_x: Optional[torch.Tensor] = None

    def to_stop(self, i: int, xnext: torch.Tensor, x: torch.Tensor,
                f: torch.Tensor, fprev: torch.Tensor) -> bool:
        """Check if the stopping conditions are met.

        Parameters
        ----------
        i: int
            The iteration number.
        xnext: torch.Tensor
            The next input.
        x: torch.Tensor
            The current input.
        f: torch.Tensor
            The current output.
        fprev: torch.Tensor
            The previous output.

        Returns
        -------
        bool
            Whether to stop the iteration.

        """
        xnorm: float = float(x.detach().norm().item())
        dxnorm: float = float((x - xnext).detach().norm().item())
        fabs: float = float(f.detach().abs().item())
        df: float = float((fprev - f).detach().abs().item())
        fval: float = float(f.detach().item())

        xtcheck = dxnorm < self.x_tol
        xrcheck = dxnorm < self.x_rtol * xnorm
        ytcheck = df < self.f_tol
        yrcheck = df < self.f_rtol * fabs
        converge = xtcheck or xrcheck or ytcheck or yrcheck
        if self.verbose:
            if i == 0:
                print("   #:             f |        dx,        df")
            if converge:
                print("Finish with convergence")
            if i == 0 or ((i + 1) % 10) == 0 or converge:
                print("%4d: %.6e | %.3e, %.3e" % (i + 1, f, dxnorm, df))

        res = (i > 0 and converge)

        # get the best values
        if not self._ever_converge and res:
            self._ever_converge = True
        if i > self._max_i:
            self._max_i = i
        if fval < self._best_f:
            self._best_f = fval
            self._best_x = x
            self._best_dxnorm = dxnorm
            self._best_df = df
        return res

    def get_best_x(self, x: torch.Tensor) -> torch.Tensor:
        """Get the best input.

        Parameters
        ----------
        x: torch.Tensor
            The current input.

        Returns
        -------
        torch.Tensor
            The best input.

        """
        # usually user set maxiter == 0 just to wrap the minimizer backprop
        if not self._ever_converge and self._max_i > -1:
            msg = (
                "The minimizer does not converge after %d iterations. "
                "Best |dx|=%.4e, |df|=%.4e, f=%.4e" %
                (self._max_i, self._best_dxnorm, self._best_df, self._best_f))
            warnings.warn(msg)
            assert isinstance(self._best_x, torch.Tensor)
            return self._best_x
        else:
            return x



import numpy as np
import tensorflow as tf

from deepchem.models import KerasModel, layers
from deepchem.models.losses import SigmoidCrossEntropy
from tensorflow.keras.layers import Input, Layer, Activation, Concatenate, Lambda


class IRVLayer(Layer):
    """ Core layer of IRV classifier, architecture described in:
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2750043/
    """

    def __init__(self, n_tasks, K, penalty, **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        K: int
            Number of nearest neighbours used in classification
        penalty: float
            Amount of penalty (l2 or l1 applied)
        """
        self.n_tasks = n_tasks
        self.K = K
        self.penalty = penalty
        super(IRVLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        self.V = tf.Variable(tf.constant([0.01, 1.]),
                             name="vote",
                             dtype=tf.float32)
        self.W = tf.Variable(tf.constant([1., 1.]), name="w", dtype=tf.float32)
        self.b = tf.Variable(tf.constant([0.01]), name="b", dtype=tf.float32)
        self.b2 = tf.Variable(tf.constant([0.01]), name="b2", dtype=tf.float32)

    def call(self, inputs):
        K = self.K
        outputs = []
        for count in range(self.n_tasks):
            # Similarity values
            similarity = inputs[:, 2 * K * count:(2 * K * count + K)]
            # Labels for all top K similar samples
            ys = tf.cast(inputs[:, (2 * K * count + K):2 * K * (count + 1)],
                         tf.int32)

            R = self.b + self.W[0] * similarity + self.W[1] * tf.constant(
                np.arange(K) + 1, dtype=tf.float32)
            R = tf.sigmoid(R)
            z = tf.reduce_sum(R * tf.gather(self.V, ys), axis=1) + self.b2
            outputs.append(tf.reshape(z, shape=[-1, 1]))
        loss = (tf.nn.l2_loss(self.W) + tf.nn.l2_loss(self.V) +
                tf.nn.l2_loss(self.b) + tf.nn.l2_loss(self.b2)) * self.penalty
        self.add_loss(loss)
        return tf.concat(outputs, axis=1)


class Slice(Layer):
    """ Choose a slice of input on the last axis given order,
    Suppose input x has two dimensions,
    output f(x) = x[:, slice_num:slice_num+1]
    """

    def __init__(self, slice_num, axis=1, **kwargs):
        """
        Parameters
        ----------
        slice_num: int
            index of slice number
        axis: int
            axis id
        """
        self.slice_num = slice_num
        self.axis = axis
        super(Slice, self).__init__(**kwargs)

    def call(self, inputs):
        slice_num = self.slice_num
        axis = self.axis
        return tf.slice(inputs, [0] * axis + [slice_num], [-1] * axis + [1])


class MultitaskIRVClassifier(KerasModel):

    def __init__(self,
                 n_tasks,
                 K=10,
                 penalty=0.0,
                 mode="classification",
                 **kwargs):
        """Initialize MultitaskIRVClassifier

        Parameters
        ----------
        n_tasks: int
            Number of tasks
        K: int
            Number of nearest neighbours used in classification
        penalty: float
            Amount of penalty (l2 or l1 applied)

        """
        self.n_tasks = n_tasks
        self.K = K
        self.n_features = 2 * self.K * self.n_tasks
        self.penalty = penalty
        mol_features = Input(shape=(self.n_features,))
        predictions = IRVLayer(self.n_tasks, self.K, self.penalty)(mol_features)
        logits = []
        outputs = []
        for task in range(self.n_tasks):
            task_output = Slice(task, 1)(predictions)
            sigmoid = Activation(tf.sigmoid)(task_output)
            logits.append(task_output)
            outputs.append(sigmoid)
        outputs = layers.Stack(axis=1)(outputs)
        outputs2 = Lambda(lambda x: 1 - x)(outputs)
        outputs = [
            Concatenate(axis=2)([outputs2, outputs]),
            logits[0] if len(logits) == 1 else Concatenate(axis=1)(logits)
        ]
        model = tf.keras.Model(inputs=[mol_features], outputs=outputs)
        super(MultitaskIRVClassifier,
              self).__init__(model,
                             SigmoidCrossEntropy(),
                             output_types=['prediction', 'loss'],
                             **kwargs)


import warnings  # noqa: E402


class TensorflowMultitaskIRVClassifier(MultitaskIRVClassifier):

    def __init__(self, *args, **kwargs):

        warnings.warn(
            "TensorflowMultitaskIRVClassifier is deprecated and has been renamed to MultitaskIRVClassifier",
            FutureWarning)

        super(TensorflowMultitaskIRVClassifier, self).__init__(*args, **kwargs)



import pytest


@pytest.mark.torch
def test_Cache():
    from deepchem.utils.cache_utils import Cache

    class A:

        def __init__(self):
            self.cache = Cache.get_dummy()

        def foo(self, x):
            return self.cache.cache("foo", lambda: x * x)

    a = A()
    assert a.foo(2) == 4


@pytest.mark.torch
def test_PrefixedCache():
    from deepchem.utils.cache_utils import Cache, _PrefixedCache
    cache = Cache.get_dummy()
    pcache = _PrefixedCache(cache, "prefix.")
    assert pcache.cache("foo", lambda: 1) == 1
    assert pcache.cache("foo", lambda: 2) == 2


@pytest.mark.torch
def test_DummyCache():
    from deepchem.utils.cache_utils import _DummyCache
    cache = _DummyCache()
    assert cache.cache("foo", lambda: 1) == 1
    assert cache.cache("foo", lambda: 2) == 2



"""
Implementation of Smiles2Vec and ChemCeption models as part of the ChemNet
transfer learning protocol.
"""

__author__ = "Vignesh Ram Somnath"
__license__ = "MIT"

import numpy as np
import tensorflow as tf

from typing import Dict
from deepchem.data.datasets import pad_batch
from deepchem.models import KerasModel
from deepchem.models.losses import L2Loss, SoftmaxCrossEntropy, SigmoidCrossEntropy
from deepchem.metrics import to_one_hot
from deepchem.models import chemnet_layers
from tensorflow.keras.layers import Input, Dense, Reshape, Softmax, Activation
from tensorflow.keras.layers import Conv1D, GRU, LSTM, Bidirectional
from tensorflow.keras.layers import GlobalAveragePooling2D

DEFAULT_INCEPTION_BLOCKS = {"A": 3, "B": 3, "C": 3}

INCEPTION_DICT = {
    "A": chemnet_layers.InceptionResnetA,
    "B": chemnet_layers.InceptionResnetB,
    "C": chemnet_layers.InceptionResnetC
}

RNN_DICT = {"GRU": GRU, "LSTM": LSTM}


class Smiles2Vec(KerasModel):
    """
    Implements the Smiles2Vec model, that learns neural representations of SMILES
    strings which can be used for downstream tasks.

    The model is based on the description in Goh et al., "SMILES2vec: An
    Interpretable General-Purpose Deep Neural Network for Predicting Chemical
    Properties" (https://arxiv.org/pdf/1712.02034.pdf). The goal here is to take
    SMILES strings as inputs, turn them into vector representations which can then
    be used in predicting molecular properties.

    The model consists of an Embedding layer that retrieves embeddings for each
    character in the SMILES string. These embeddings are learnt jointly with the
    rest of the model. The output from the embedding layer is a tensor of shape
    (batch_size, seq_len, embedding_dim). This tensor can optionally be fed
    through a 1D convolutional layer, before being passed to a series of RNN cells
    (optionally bidirectional). The final output from the RNN cells aims
    to have learnt the temporal dependencies in the SMILES string, and in turn
    information about the structure of the molecule, which is then used for
    molecular property prediction.

    In the paper, the authors also train an explanation mask to endow the model
    with interpretability and gain insights into its decision making. This segment
    is currently not a part of this implementation as this was
    developed for the purpose of investigating a transfer learning protocol,
    ChemNet (which can be found at https://arxiv.org/abs/1712.02734).
    """

    def __init__(self,
                 char_to_idx,
                 n_tasks=10,
                 max_seq_len=270,
                 embedding_dim=50,
                 n_classes=2,
                 use_bidir=True,
                 use_conv=True,
                 filters=192,
                 kernel_size=3,
                 strides=1,
                 rnn_sizes=[224, 384],
                 rnn_types=["GRU", "GRU"],
                 mode="regression",
                 **kwargs):
        """
        Parameters
        ----------
        char_to_idx: dict,
            char_to_idx contains character to index mapping for SMILES characters
        embedding_dim: int, default 50
            Size of character embeddings used.
        use_bidir: bool, default True
            Whether to use BiDirectional RNN Cells
        use_conv: bool, default True
            Whether to use a 1D conv-layer
        kernel_size: int, default 3
            Kernel size for convolutions
        filters: int, default 192
            Number of filters
        strides: int, default 1
            Strides used in convolution
        rnn_sizes: list[int], default [224, 384]
            Number of hidden units in the RNN cells
        mode: str, default regression
            Whether to use model for regression or classification
        """

        self.char_to_idx = char_to_idx
        self.n_classes = n_classes
        self.max_seq_len = max_seq_len
        self.embedding_dim = embedding_dim
        self.use_bidir = use_bidir
        self.use_conv = use_conv
        if use_conv:
            self.kernel_size = kernel_size
            self.filters = filters
            self.strides = strides
        self.rnn_types = rnn_types
        self.rnn_sizes = rnn_sizes
        assert len(rnn_sizes) == len(
            rnn_types), "Should have same number of hidden units as RNNs"
        self.n_tasks = n_tasks
        self.mode = mode

        model, loss, output_types = self._build_graph()
        super(Smiles2Vec, self).__init__(model=model,
                                         loss=loss,
                                         output_types=output_types,
                                         **kwargs)

    def _build_graph(self):
        """Build the model."""
        smiles_seqs = Input(dtype=tf.int32,
                            shape=(self.max_seq_len,),
                            name='Input')
        rnn_input = tf.keras.layers.Embedding(
            input_dim=len(self.char_to_idx),
            output_dim=self.embedding_dim)(smiles_seqs)

        if self.use_conv:
            rnn_input = Conv1D(filters=self.filters,
                               kernel_size=self.kernel_size,
                               strides=self.strides,
                               activation=tf.nn.relu,
                               name='Conv1D')(rnn_input)

        rnn_embeddings = rnn_input
        for idx, rnn_type in enumerate(self.rnn_types[:-1]):
            rnn_layer = RNN_DICT[rnn_type]
            layer = rnn_layer(units=self.rnn_sizes[idx], return_sequences=True)
            if self.use_bidir:
                layer = Bidirectional(layer)

            rnn_embeddings = layer(rnn_embeddings)

        # Last layer sequences not returned.
        layer = RNN_DICT[self.rnn_types[-1]](units=self.rnn_sizes[-1])
        if self.use_bidir:
            layer = Bidirectional(layer)
        rnn_embeddings = layer(rnn_embeddings)

        if self.mode == "classification":
            logits = Dense(self.n_tasks * self.n_classes)(rnn_embeddings)
            logits = Reshape((self.n_tasks, self.n_classes))(logits)
            if self.n_classes == 2:
                output = Activation(activation='sigmoid')(logits)
                loss = SigmoidCrossEntropy()
            else:
                output = Softmax()(logits)
                loss = SoftmaxCrossEntropy()
            outputs = [output, logits]
            output_types = ['prediction', 'loss']

        else:
            output = Dense(self.n_tasks * 1, name='Dense')(rnn_embeddings)
            output = Reshape((self.n_tasks, 1), name='Reshape')(output)
            outputs = [output]
            output_types = ['prediction']
            loss = L2Loss()

        model = tf.keras.Model(inputs=[smiles_seqs], outputs=outputs)
        return model, loss, output_types

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True):
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if self.mode == 'classification':
                    y_b = to_one_hot(y_b.flatten(), self.n_classes).reshape(
                        -1, self.n_tasks, self.n_classes)
                yield ([X_b], [y_b], [w_b])


class ChemCeption(KerasModel):
    """
    Implements the ChemCeption model that leverages the representational capacities
    of convolutional neural networks (CNNs) to predict molecular properties.

    The model is based on the description in Goh et al., "Chemception: A Deep
    Neural Network with Minimal Chemistry Knowledge Matches the Performance of
    Expert-developed QSAR/QSPR Models" (https://arxiv.org/pdf/1706.06689.pdf).
    The authors use an image based representation of the molecule, where pixels
    encode different atomic and bond properties. More details on the image repres-
    entations can be found at https://arxiv.org/abs/1710.02238

    The model consists of a Stem Layer that reduces the image resolution for the
    layers to follow. The output of the Stem Layer is followed by a series of
    Inception-Resnet blocks & a Reduction layer. Layers in the Inception-Resnet
    blocks process image tensors at multiple resolutions and use a ResNet style
    skip-connection, combining features from different resolutions. The Reduction
    layers reduce the spatial extent of the image by max-pooling and 2-strided
    convolutions. More details on these layers can be found in the ChemCeption
    paper referenced above. The output of the final Reduction layer is subject to
    a Global Average Pooling, and a fully-connected layer maps the features to
    downstream outputs.

    In the ChemCeption paper, the authors perform real-time image augmentation by
    rotating images between 0 to 180 degrees. This can be done during model
    training by setting the augment argument to True.
    """

    def __init__(self,
                 img_spec: str = "std",
                 img_size: int = 80,
                 base_filters: int = 16,
                 inception_blocks: Dict = DEFAULT_INCEPTION_BLOCKS,
                 n_tasks: int = 10,
                 n_classes: int = 2,
                 augment: bool = False,
                 mode: str = "regression",
                 **kwargs):
        """
        Parameters
        ----------
        img_spec: str, default std
            Image specification used
        img_size: int, default 80
            Image size used
        base_filters: int, default 16
            Base filters used for the different inception and reduction layers
        inception_blocks: dict,
            Dictionary containing number of blocks for every inception layer
        n_tasks: int, default 10
            Number of classification or regression tasks
        n_classes: int, default 2
            Number of classes (used only for classification)
        augment: bool, default False
            Whether to augment images
        mode: str, default regression
            Whether the model is used for regression or classification
        """
        if img_spec == "engd":
            self.input_shape = (img_size, img_size, 4)
        elif img_spec == "std":
            self.input_shape = (img_size, img_size, 1)
        self.base_filters = base_filters
        self.inception_blocks = inception_blocks
        self.n_tasks = n_tasks
        self.n_classes = n_classes
        self.mode = mode
        self.augment = augment

        model, loss, output_types = self._build_graph()
        super(ChemCeption, self).__init__(model=model,
                                          loss=loss,
                                          output_types=output_types,
                                          **kwargs)

    def _build_graph(self):
        smile_images = Input(shape=self.input_shape)
        stem = chemnet_layers.Stem(self.base_filters)(smile_images)

        inceptionA_out = self.build_inception_module(inputs=stem, type="A")
        reductionA_out = chemnet_layers.ReductionA(
            self.base_filters)(inceptionA_out)

        inceptionB_out = self.build_inception_module(inputs=reductionA_out,
                                                     type="B")
        reductionB_out = chemnet_layers.ReductionB(
            self.base_filters)(inceptionB_out)

        inceptionC_out = self.build_inception_module(inputs=reductionB_out,
                                                     type="C")
        avg_pooling_out = GlobalAveragePooling2D()(inceptionC_out)

        if self.mode == "classification":
            logits = Dense(self.n_tasks * self.n_classes)(avg_pooling_out)
            logits = Reshape((self.n_tasks, self.n_classes))(logits)
            if self.n_classes == 2:
                output = Activation(activation='sigmoid')(logits)
                loss = SigmoidCrossEntropy()
            else:
                output = Softmax()(logits)
                loss = SoftmaxCrossEntropy()
            outputs = [output, logits]
            output_types = ['prediction', 'loss']

        else:
            output = Dense(self.n_tasks * 1)(avg_pooling_out)
            output = Reshape((self.n_tasks, 1))(output)
            outputs = [output]
            output_types = ['prediction']
            loss = L2Loss()

        model = tf.keras.Model(inputs=[smile_images], outputs=outputs)
        return model, loss, output_types

    def build_inception_module(self, inputs, type="A"):
        """Inception module is a series of inception layers of similar type. This
        function builds that."""
        num_blocks = self.inception_blocks[type]
        inception_layer = INCEPTION_DICT[type]
        output = inputs
        for block in range(num_blocks):
            output = inception_layer(self.base_filters,
                                     int(inputs.shape[-1]))(output)
        return output

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True):
        for epoch in range(epochs):
            if mode == "predict" or (not self.augment):
                for (X_b, y_b, w_b,
                     ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                                   deterministic=deterministic,
                                                   pad_batches=pad_batches):
                    if self.mode == 'classification':
                        y_b = to_one_hot(y_b.flatten(), self.n_classes).reshape(
                            -1, self.n_tasks, self.n_classes)
                    yield ([X_b], [y_b], [w_b])

            else:
                if not pad_batches:
                    n_samples = dataset.X.shape[0]
                else:
                    n_samples = dataset.X.shape[0] + (
                        self.batch_size -
                        (dataset.X.shape[0] % self.batch_size))

                n_batches = 0
                image_data_generator = tf.keras.preprocessing.image.ImageDataGenerator(
                    rotation_range=180)
                for (X_b, y_b, w_b) in image_data_generator.flow(
                        dataset.X,
                        dataset.y,
                        sample_weight=dataset.w,
                        shuffle=not deterministic,
                        batch_size=self.batch_size):
                    if pad_batches:
                        ids_b = np.arange(X_b.shape[0])
                        X_b, y_b, w_b, _ = pad_batch(self.batch_size, X_b, y_b,
                                                     w_b, ids_b)
                    n_batches += 1
                    if n_batches > n_samples / self.batch_size:
                        # This is needed because ImageDataGenerator does infinite looping
                        break
                    if self.mode == "classification":
                        y_b = to_one_hot(y_b.flatten(), self.n_classes).reshape(
                            -1, self.n_tasks, self.n_classes)
                    yield ([X_b], [y_b], [w_b])



"""
Conformer generation.
"""

import numpy as np
from typing import Any, List, Optional
from deepchem.utils.typing import RDKitMol


class ConformerGenerator(object):
    """
    Generate molecule conformers.

    Notes
    -----
    Procedure
    1. Generate a pool of conformers.
    2. Minimize conformers.
    3. Prune conformers using an RMSD threshold.

    Note that pruning is done _after_ minimization, which differs from the
    protocol described in the references [1]_ [2]_.

    References
    ----------
    .. [1] http://rdkit.org/docs/GettingStartedInPython.html#working-with-3d-molecules
    .. [2] http://pubs.acs.org/doi/full/10.1021/ci2004658

    Notes
    -----
    This class requires RDKit to be installed.
    """

    def __init__(self,
                 max_conformers: int = 1,
                 rmsd_threshold: float = 0.5,
                 force_field: str = 'uff',
                 pool_multiplier: int = 10):
        """
        Parameters
        ----------
        max_conformers: int, optional (default 1)
            Maximum number of conformers to generate (after pruning).
        rmsd_threshold: float, optional (default 0.5)
            RMSD threshold for pruning conformers. If None or negative, no
            pruning is performed.
        force_field: str, optional (default 'uff')
            Force field to use for conformer energy calculation and
            minimization. Options are 'uff', 'mmff94', and 'mmff94s'.
        pool_multiplier: int, optional (default 10)
            Factor to multiply by max_conformers to generate the initial
            conformer pool. Since conformers are pruned after energy
            minimization, increasing the size of the pool increases the chance
            of identifying max_conformers unique conformers.
        """
        self.max_conformers = max_conformers
        if rmsd_threshold is None or rmsd_threshold < 0:
            rmsd_threshold = -1.
        self.rmsd_threshold = rmsd_threshold
        self.force_field = force_field
        self.pool_multiplier = pool_multiplier

    def __call__(self, mol: RDKitMol) -> RDKitMol:
        """
        Generate conformers for a molecule.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        mol: rdkit.Chem.rdchem.Mol
            A new RDKit Mol object containing the chosen conformers, sorted by
            increasing energy.
        """
        return self.generate_conformers(mol)

    def generate_conformers(self, mol: RDKitMol) -> RDKitMol:
        """
        Generate conformers for a molecule.

        This function returns a copy of the original molecule with embedded
        conformers.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        mol: rdkit.Chem.rdchem.Mol
            A new RDKit Mol object containing the chosen conformers, sorted by
            increasing energy.
        """

        # initial embedding
        mol = self.embed_molecule(mol)
        if not mol.GetNumConformers():
            msg = 'No conformers generated for molecule'
            if mol.HasProp('_Name'):
                name = mol.GetProp('_Name')
                msg += ' "{}".'.format(name)
            else:
                msg += '.'
            raise RuntimeError(msg)

        # minimization and pruning
        self.minimize_conformers(mol)
        mol = self.prune_conformers(mol)

        return mol

    def embed_molecule(self, mol: RDKitMol) -> RDKitMol:
        """
        Generate conformers, possibly with pruning.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object with embedded multiple conformers.
        """
        try:
            from rdkit import Chem
            from rdkit.Chem import AllChem
        except ModuleNotFoundError:
            raise ImportError("This function requires RDKit to be installed.")

        mol = Chem.AddHs(mol)  # add hydrogens
        n_confs = self.max_conformers * self.pool_multiplier
        AllChem.EmbedMultipleConfs(mol, numConfs=n_confs, pruneRmsThresh=-1.)
        return mol

    def get_molecule_force_field(self,
                                 mol: RDKitMol,
                                 conf_id: Optional[int] = None,
                                 **kwargs) -> Any:
        """
        Get a force field for a molecule.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object with embedded conformers.
        conf_id: int, optional
            ID of the conformer to associate with the force field.
        kwargs: dict, optional
            Keyword arguments for force field constructor.

        Returns
        -------
        ff: rdkit.ForceField.rdForceField.ForceField
            RDKit force field instance for a molecule.
        """
        try:
            from rdkit.Chem import AllChem
        except ModuleNotFoundError:
            raise ImportError("This function requires RDKit to be installed.")

        if self.force_field == 'uff':
            ff = AllChem.UFFGetMoleculeForceField(mol, confId=conf_id, **kwargs)
        elif self.force_field.startswith('mmff'):
            AllChem.MMFFSanitizeMolecule(mol)
            mmff_props = AllChem.MMFFGetMoleculeProperties(
                mol, mmffVariant=self.force_field)
            ff = AllChem.MMFFGetMoleculeForceField(mol,
                                                   mmff_props,
                                                   confId=conf_id,
                                                   **kwargs)
        else:
            raise ValueError("Invalid force_field " +
                             "'{}'.".format(self.force_field))
        return ff

    def minimize_conformers(self, mol: RDKitMol) -> None:
        """
        Minimize molecule conformers.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object with embedded conformers.
        """
        for conf in mol.GetConformers():
            ff = self.get_molecule_force_field(mol, conf_id=conf.GetId())
            ff.Minimize()

    def get_conformer_energies(self, mol: RDKitMol) -> np.ndarray:
        """
        Calculate conformer energies.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object with embedded conformers.

        Returns
        -------
        energies : np.ndarray
            Minimized conformer energies.
        """
        energies = []
        for conf in mol.GetConformers():
            ff = self.get_molecule_force_field(mol, conf_id=conf.GetId())
            energy = ff.CalcEnergy()
            energies.append(energy)
        return np.asarray(energies, dtype=float)

    def prune_conformers(self, mol: RDKitMol) -> RDKitMol:
        """
        Prune conformers from a molecule using an RMSD threshold, starting
        with the lowest energy conformer.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        new_mol: rdkit.Chem.rdchem.Mol
            A new rdkit.Chem.rdchem.Mol containing the chosen conformers, sorted by
            increasing energy.
        """
        try:
            from rdkit import Chem
        except ModuleNotFoundError:
            raise ImportError("This function requires RDKit to be installed.")

        if self.rmsd_threshold < 0 or mol.GetNumConformers() <= 1:
            return mol
        energies = self.get_conformer_energies(mol)
        rmsd = self.get_conformer_rmsd(mol)

        sort = np.argsort(energies)  # sort by increasing energy
        keep: List[float] = []  # always keep lowest-energy conformer
        discard = []
        for i in sort:
            # always keep lowest-energy conformer
            if len(keep) == 0:
                keep.append(i)
                continue

            # discard conformers after max_conformers is reached
            if len(keep) >= self.max_conformers:
                discard.append(i)
                continue

            # get RMSD to selected conformers
            this_rmsd = rmsd[i][np.asarray(keep, dtype=int)]

            # discard conformers within the RMSD threshold
            if np.all(this_rmsd >= self.rmsd_threshold):
                keep.append(i)
            else:
                discard.append(i)

        # create a new molecule to hold the chosen conformers
        # this ensures proper conformer IDs and energy-based ordering
        new_mol = Chem.Mol(mol)
        new_mol.RemoveAllConformers()
        conf_ids = [conf.GetId() for conf in mol.GetConformers()]
        for i in keep:
            conf = mol.GetConformer(conf_ids[i])
            new_mol.AddConformer(conf, assignId=True)
        return new_mol

    @staticmethod
    def get_conformer_rmsd(mol: RDKitMol) -> np.ndarray:
        """
        Calculate conformer-conformer RMSD.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        rmsd: np.ndarray
            A conformer-conformer RMSD value. The shape is `(NumConformers, NumConformers)`
        """
        try:
            from rdkit.Chem import AllChem
        except ModuleNotFoundError:
            raise ImportError("This function requires RDKit to be installed.")

        rmsd = np.zeros((mol.GetNumConformers(), mol.GetNumConformers()),
                        dtype=float)
        for i, ref_conf in enumerate(mol.GetConformers()):
            for j, fit_conf in enumerate(mol.GetConformers()):
                if i >= j:
                    continue
                rmsd[i, j] = AllChem.GetBestRMS(mol, mol, ref_conf.GetId(),
                                                fit_conf.GetId())
                rmsd[j, i] = rmsd[i, j]
        return rmsd



"""
Various utilities around voxel grids.
"""
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import numpy as np

from deepchem.utils.noncovalent_utils import compute_pi_stack

logger = logging.getLogger(__name__)


def convert_atom_to_voxel(coordinates: np.ndarray, atom_index: int,
                          box_width: float, voxel_width: float) -> np.ndarray:
    """Converts atom coordinates to an i,j,k grid index.

    This function offsets molecular atom coordinates by
    (box_width/2, box_width/2, box_width/2) and then divides by
    voxel_width to compute the voxel indices.

    Parameters
    -----------
    coordinates: np.ndarray
        Array with coordinates of all atoms in the molecule, shape (N, 3).
    atom_index: int
        Index of an atom in the molecule.
    box_width: float
        Size of the box in Angstroms.
    voxel_width: float
        Size of a voxel in Angstroms

    Returns
    -------
    indices: np.ndarray
        A 1D numpy array of length 3 with `[i, j, k]`, the voxel coordinates
        of specified atom.
    """

    indices = np.floor(
        (coordinates[atom_index] + box_width / 2.0) / voxel_width).astype(int)

    return indices


def convert_atom_pair_to_voxel(coordinates_tuple: Tuple[np.ndarray, np.ndarray],
                               atom_index_pair: Tuple[int,
                                                      int], box_width: float,
                               voxel_width: float) -> np.ndarray:
    """Converts a pair of atoms to i,j,k grid indexes.


    Parameters
    ----------
    coordinates_tuple: Tuple[np.ndarray, np.ndarray]
        A tuple containing two molecular coordinate arrays of shapes `(N, 3)` and `(M, 3)`.
    atom_index_pair: Tuple[int, int]
        A tuple of indices for the atoms in the two molecules.
    box_width: float
        Size of the box in Angstroms.
    voxel_width: float
        Size of a voxel in Angstroms

    Returns
    -------
    indices_list: np.ndarray
        A numpy array of shape `(2, 3)`, where `3` is `[i, j, k]` of the
        voxel coordinates of specified atom.
    """

    indices_list = []
    for coordinates, atom_index in zip(coordinates_tuple, atom_index_pair):
        indices_list.append(
            convert_atom_to_voxel(coordinates, atom_index, box_width,
                                  voxel_width))
    return np.array(indices_list)


def voxelize(get_voxels: Callable[..., Any],
             coordinates: Any,
             box_width: float = 16.0,
             voxel_width: float = 1.0,
             hash_function: Optional[Callable[..., Any]] = None,
             feature_dict: Optional[Dict[Any, Any]] = None,
             feature_list: Optional[List[Union[int, Tuple[int]]]] = None,
             nb_channel: int = 16,
             dtype: str = 'int') -> np.ndarray:
    """Helper function to voxelize inputs.

    This helper function helps convert a hash function which
    specifies spatial features of a molecular complex into a voxel
    tensor. This utility is used by various featurizers that generate
    voxel grids.

    Parameters
    ----------
    get_voxels: Function
        Function that voxelizes inputs
    coordinates: Any
        Contains the 3D coordinates of a molecular system.  This should have
        whatever type get_voxels() expects as its first argument.
    box_width: float, optional (default 16.0)
        Size of a box in which voxel features are calculated. Box
        is centered on a ligand centroid.
    voxel_width: float, optional (default 1.0)
        Size of a 3D voxel in a grid in Angstroms.
    hash_function: Function
        Used to map feature choices to voxel channels.
    feature_dict: Dict, optional (default None)
        Keys are atom indices or tuples of atom indices, the values are
        computed features. If `hash_function is not None`, then the values
        are hashed using the hash function into `[0, nb_channels)` and
        this channel at the voxel for the given key is incremented by `1`
        for each dictionary entry. If `hash_function is None`, then the
        value must be a vector of size `(n_channels,)` which is added to
        the existing channel values at that voxel grid.
    feature_list: List, optional (default None)
        List of atom indices or tuples of atom indices. This can only be
        used if `nb_channel==1`. Increments the voxels corresponding to
        these indices by `1` for each entry.
    nb_channel: int, , optional (default 16)
        The number of feature channels computed per voxel. Should
        be a power of 2.
    dtype: str ('int' or 'float'), optional (default 'int')
        The type of the numpy ndarray created to hold features.

    Returns
    -------
    feature_tensor: np.ndarray
        The voxel of the input with the shape
        `(voxels_per_edge, voxels_per_edge, voxels_per_edge, nb_channel)`.
    """
    # Number of voxels per one edge of box to voxelize.
    voxels_per_edge = int(box_width / voxel_width)
    if dtype == "int":
        feature_tensor = np.zeros(
            (voxels_per_edge, voxels_per_edge, voxels_per_edge, nb_channel),
            dtype=np.int8)
    else:
        feature_tensor = np.zeros(
            (voxels_per_edge, voxels_per_edge, voxels_per_edge, nb_channel),
            dtype=np.float16)
    if feature_dict is not None:
        for key, features in feature_dict.items():
            voxels = get_voxels(coordinates, key, box_width, voxel_width)
            if len(voxels.shape) == 1:
                voxels = np.expand_dims(voxels, axis=0)
            for voxel in voxels:
                if ((voxel >= 0) & (voxel < voxels_per_edge)).all():
                    if hash_function is not None:
                        feature_tensor[
                            voxel[0], voxel[1], voxel[2],
                            hash_function(features, nb_channel)] += 1.0
                    else:
                        feature_tensor[voxel[0], voxel[1], voxel[2],
                                       0] += features
    elif feature_list is not None:
        for key in feature_list:
            voxels = get_voxels(coordinates, key, box_width, voxel_width)
            for voxel in voxels:
                if ((voxel >= 0) & (voxel < voxels_per_edge)).all():
                    feature_tensor[voxel[0], voxel[1], voxel[2], 0] += 1.0

    return feature_tensor


def voxelize_pi_stack(prot_xyz, prot_rdk, lig_xyz, lig_rdk, distances,
                      pi_stack_dist_cutoff, pi_stack_angle_cutoff, box_width,
                      voxel_width):
    protein_pi_t, protein_pi_parallel, ligand_pi_t, ligand_pi_parallel = (
        compute_pi_stack(prot_rdk,
                         lig_rdk,
                         distances,
                         dist_cutoff=pi_stack_dist_cutoff,
                         angle_cutoff=pi_stack_angle_cutoff))
    pi_parallel_tensor = voxelize(
        convert_atom_to_voxel,
        prot_xyz,
        box_width=box_width,
        voxel_width=voxel_width,
        feature_dict=protein_pi_parallel,
        nb_channel=1,
    )

    pi_parallel_tensor += voxelize(
        convert_atom_to_voxel,
        lig_xyz,
        box_width=box_width,
        voxel_width=voxel_width,
        feature_dict=ligand_pi_parallel,
        nb_channel=1,
    )

    pi_t_tensor = voxelize(
        convert_atom_to_voxel,
        prot_xyz,
        box_width=box_width,
        voxel_width=voxel_width,
        feature_dict=protein_pi_t,
        nb_channel=1,
    )

    pi_t_tensor += voxelize(
        convert_atom_to_voxel,
        lig_xyz,
        box_width=box_width,
        voxel_width=voxel_width,
        feature_dict=ligand_pi_t,
        nb_channel=1,
    )

    return [pi_parallel_tensor, pi_t_tensor]



from __future__ import annotations
import torch
from dataclasses import dataclass
from typing import Optional, Union, Dict
from deepchem.utils.dft_utils.hamilton.intor.utils import NDIM


@dataclass
class PBCIntOption:
    """PBCIntOption is a class that contains parameters for the PBC integrals.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import PBCIntOption
    >>> pbc = PBCIntOption()
    >>> pbc.get_default()
    PBCIntOption(precision=1e-08, kpt_diff_tol=1e-06)

    Attributes
    ----------
    precision: float (default 1e-8)
        Precision of the integral to limit the lattice sum.
    kpt_diff_tol: float (default 1e-6)
        Difference between k-points to be regarded as the same.

    """
    precision: float = 1e-8
    kpt_diff_tol: float = 1e-6

    @staticmethod
    def get_default(
        lattsum_opt: Optional[Union[PBCIntOption,
                                    Dict]] = None) -> PBCIntOption:
        """Get the default PBCIntOption object.

        Parameters
        ----------
        lattsum_opt: Optional[Union[PBCIntOption, Dict]]
            The lattice sum option. If it is a dictionary, then it will be
            converted to a PBCIntOption object. If it is None, then just use
            the default value of PBCIntOption.

        Returns
        -------
        PBCIntOption
            The default PBCIntOption object.

        """
        if lattsum_opt is None:
            return PBCIntOption()
        elif isinstance(lattsum_opt, dict):
            return PBCIntOption(**lattsum_opt)
        else:
            return lattsum_opt


# helper functions
def get_default_options(options: Optional[PBCIntOption] = None) -> PBCIntOption:
    """if options is None, then set the default option.
    otherwise, just return the input options.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import get_default_options
    >>> get_default_options()
    PBCIntOption(precision=1e-08, kpt_diff_tol=1e-06)

    Parameters
    ----------
    options: Optional[PBCIntOption]
        Input options object

    Returns
    -------
    PBCIntOption
        The options object

    """
    if options is None:
        options1 = PBCIntOption()
    else:
        options1 = options
    return options1


def get_default_kpts(kpts: Optional[torch.Tensor], dtype: torch.dtype,
                     device: torch.device) -> torch.Tensor:
    """if kpts is None, then set the default kpts (k = zeros)
    otherwise, just return the input kpts in the correct dtype and device

    Examples
    --------
    >>> from deepchem.utils.dft_utils import get_default_kpts
    >>> get_default_kpts(torch.tensor([[1, 1, 1]]), torch.float64, 'cpu')
    tensor([[1., 1., 1.]], dtype=torch.float64)

    Parameters
    ----------
    kpts: Optional[torch.Tensor]
        Input k-points
    dtype: torch.dtype
        The dtype of the kpts
    device: torch.device
        Device on which the tensord are located. Ex: cuda, cpu

    Returns
    -------
    torch.Tensor
        Default k-points

    """
    if kpts is None:
        kpts1 = torch.zeros((1, NDIM), dtype=dtype, device=device)
    else:
        kpts1 = kpts.to(dtype).to(device)
        assert kpts1.ndim == 2
        assert kpts1.shape[-1] == NDIM
    return kpts1



"""Utility functions for working with PyTorch."""
import scipy
import torch
import numpy as np
from typing import Any, Callable, Sequence, Union, List, Generator, Tuple


def get_activation(fn: Union[Callable, str]):
    """Get a PyTorch activation function, specified either directly or as a string.

    This function simplifies allowing users to specify activation functions by name.
    If a function is provided, it is simply returned unchanged.  If a string is provided,
    the corresponding function in torch.nn.functional is returned.
    """
    if isinstance(fn, str):
        return getattr(torch.nn.functional, fn)
    return fn


def unsorted_segment_sum(data: torch.Tensor, segment_ids: torch.Tensor,
                         num_segments: int) -> torch.Tensor:
    """Computes the sum along segments of a tensor. Analogous to tf.unsorted_segment_sum.

    Parameters
    ----------
    data: torch.Tensor
        A tensor whose segments are to be summed.
    segment_ids: torch.Tensor
        The segment indices tensor.
    num_segments: int
        The number of segments.

    Returns
    -------
    tensor: torch.Tensor

    Examples
    --------
    >>> segment_ids = torch.Tensor([0, 1, 0]).to(torch.int64)
    >>> data = torch.Tensor([[1, 2, 3, 4], [5, 6, 7, 8], [4, 3, 2, 1]])
    >>> num_segments = 2
    >>> result = unsorted_segment_sum(data=data,
    ...                               segment_ids=segment_ids,
    ...                               num_segments=num_segments)
    >>> data.shape[0]
    3
    >>> segment_ids.shape[0]
    3
    >>> len(segment_ids.shape)
    1
    >>> result
    tensor([[5., 5., 5., 5.],
            [5., 6., 7., 8.]])

    """

    if len(segment_ids.shape) != 1:
        raise AssertionError("segment_ids have be a 1-D tensor")

    if data.shape[0] != segment_ids.shape[0]:
        raise AssertionError(
            "segment_ids should be the same size as dimension 0 of input.")

    s = torch.prod(torch.tensor(data.shape[1:])).long()
    segment_ids = segment_ids.repeat_interleave(s).view(segment_ids.shape[0],
                                                        *data.shape[1:])

    # data.shape and segment_ids.shape should be equal
    assert data.shape == segment_ids.shape
    shape: List[int] = [num_segments] + list(data.shape[1:])
    tensor: torch.Tensor = torch.zeros(*shape).scatter_add(
        0, segment_ids, data.float())
    tensor = tensor.type(data.dtype)
    return tensor


def segment_sum(data: torch.Tensor, segment_ids: torch.Tensor) -> torch.Tensor:
    """ This function computes the sum of values along segments within a tensor. It is useful when you have a tensor with segment IDs and you want to compute the sum of values for each segment.
    This function is analogous to tf.segment_sum. (https://www.tensorflow.org/api_docs/python/tf/math/segment_sum).

    Parameters
    ----------
    data: torch.Tensor
        A pytorch tensor containing the values to be summed. It can have any shape, but its rank (number of dimensions) should be at least 1.
    segment_ids: torch.Tensor
        A 1-D tensor containing the indices for the segmentation. The segments can be any non-negative integer values, but they must be sorted in non-decreasing order.

    Returns
    -------
    out_tensor: torch.Tensor
        Tensor with the same shape as data, where each value corresponds to the sum of values within the corresponding segment.

    Examples
    --------
    >>> data = torch.Tensor([[1, 2, 3, 4], [4, 3, 2, 1], [5, 6, 7, 8]])
    >>> segment_ids = torch.Tensor([0, 0, 1]).to(torch.int64)
    >>> result = segment_sum(data=data, segment_ids=segment_ids)
    >>> data.shape[0]
    3
    >>> segment_ids.shape[0]
    3
    >>> len(segment_ids.shape)
    1
    >>> result
    tensor([[5., 5., 5., 5.],
            [5., 6., 7., 8.]])

    """
    if not all(segment_ids[i] <= segment_ids[i + 1]
               for i in range(len(segment_ids) - 1)):
        raise AssertionError("elements of segment_ids must be sorted")

    if len(segment_ids.shape) != 1:
        raise AssertionError("segment_ids have be a 1-D tensor")

    if data.shape[0] != segment_ids.shape[0]:
        raise AssertionError(
            "segment_ids should be the same size as dimension 0 of input.")

    num_segments = len(torch.unique(segment_ids))
    out_tensor = unsorted_segment_sum(data, segment_ids, num_segments)

    return out_tensor


def chunkify(a: torch.Tensor, dim: int, maxnumel: int) -> \
        Generator[Tuple[torch.Tensor, int, int], None, None]:
    """Splits the tensor `a` into several chunks of size `maxnumel` along the
    dimension given by `dim`.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.pytorch_utils import chunkify
    >>> a = torch.arange(10)
    >>> for chunk, istart, iend in chunkify(a, 0, 3):
    ...     print(chunk, istart, iend)
    tensor([0, 1, 2]) 0 3
    tensor([3, 4, 5]) 3 6
    tensor([6, 7, 8]) 6 9
    tensor([9]) 9 12

    Parameters
    ----------
    a: torch.Tensor
        The big tensor to be splitted into chunks.
    dim: int
        The dimension where the tensor would be splitted.
    maxnumel: int
        Maximum number of elements in a chunk.

    Returns
    -------
    chunks: Generator[Tuple[torch.Tensor, int, int], None, None]
        A generator that yields a tuple of three elements: the chunk tensor, the
        starting index of the chunk and the ending index of the chunk.

    """
    dim = a.ndim + dim if dim < 0 else dim

    numel = a.numel()
    dimnumel = a.shape[dim]
    nondimnumel = numel // dimnumel
    if maxnumel < nondimnumel:
        msg = "Cannot split the tensor of shape %s along dimension %s with maxnumel %d" % \
              (a.shape, dim, maxnumel)
        raise RuntimeError(msg)

    csize = min(maxnumel // nondimnumel, dimnumel)
    ioffset = 0
    lslice = (slice(None, None, None),) * dim
    rslice = (slice(None, None, None),) * (a.ndim - dim - 1)
    while ioffset < dimnumel:
        iend = ioffset + csize
        chunks = a[(lslice + (slice(ioffset, iend, None),) +
                    rslice)], ioffset, iend
        yield chunks
        ioffset = iend


def get_memory(a: torch.Tensor) -> int:
    """Returns the size of the tensor in bytes.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.pytorch_utils import get_memory
    >>> a = torch.randn(100, 100, dtype=torch.float64)
    >>> get_memory(a)
    80000

    Parameters
    ----------
    a: torch.Tensor
        The tensor to be measured.

    Returns
    -------
    size: int
        The size of the tensor in bytes.

    """
    size = a.numel() * a.element_size()
    return size


def gaussian_integral(
        n: int, alpha: Union[float,
                             torch.Tensor]) -> Union[float, torch.Tensor]:
    """Performs the gaussian integration.

    Examples
    --------
    >>> gaussian_integral(5, 1.0)
    1.0

    Parameters
    ----------
    n: int
        The order of the integral
    alpha: Union[float, torch.Tensor]
        The parameter of the gaussian

    Returns
    -------
    Union[float, torch.Tensor]
        The value of the integral

    """
    n1 = (n + 1) * 0.5
    return scipy.special.gamma(n1) / (2 * alpha**n1)


class TensorNonTensorSeparator(object):
    """
    Class that provides function to separate/combine tensors and nontensors
    parameters.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.pytorch_utils import TensorNonTensorSeparator
    >>> a = torch.tensor([1.,2,3])
    >>> b = 4.
    >>> c = torch.tensor([5.,6,7], requires_grad=True)
    >>> params = [a, b, c]
    >>> separator = TensorNonTensorSeparator(params)
    >>> tensor_params = separator.get_tensor_params()
    >>> tensor_params
    [tensor([5., 6., 7.], requires_grad=True)]

    """

    def __init__(self, params: Sequence, varonly: bool = True):
        """Initialize the TensorNonTensorSeparator.

        Parameters
        ----------
        params: Sequence
            A list of tensor or non-tensor parameters.
        varonly: bool
            If True, only tensor parameters with requires_grad=True will be
            returned. Otherwise, all tensor parameters will be returned.

        """
        self.tensor_idxs = []
        self.tensor_params = []
        self.nontensor_idxs = []
        self.nontensor_params = []
        self.nparams = len(params)
        for (i, p) in enumerate(params):
            if isinstance(p, torch.Tensor) and ((varonly and p.requires_grad) or
                                                (not varonly)):
                self.tensor_idxs.append(i)
                self.tensor_params.append(p)
            else:
                self.nontensor_idxs.append(i)
                self.nontensor_params.append(p)
        self.alltensors = len(self.tensor_idxs) == self.nparams

    def get_tensor_params(self):
        """Returns a list of tensor parameters.

        Returns
        -------
        List[torch.Tensor]
            A list of tensor parameters.

        """
        return self.tensor_params

    def ntensors(self):
        """Returns the number of tensor parameters.

        Returns
        -------
        int
            The number of tensor parameters.

        """
        return len(self.tensor_idxs)

    def nnontensors(self):
        """Returns the number of nontensor parameters.

        Returns
        -------
        int
            The number of nontensor parameters.

        """
        return len(self.nontensor_idxs)

    def reconstruct_params(self, tensor_params, nontensor_params=None):
        """Reconstruct the parameters from tensor and nontensor parameters.

        Parameters
        ----------
        tensor_params: List[torch.Tensor]
            A list of tensor parameters.
        nontensor_params: Optional[List]
            A list of nontensor parameters. If None, the original nontensor
            parameters will be used.

        Returns
        -------
        List
            A list of parameters.

        """
        if nontensor_params is None:
            nontensor_params = self.nontensor_params
        if len(tensor_params) + len(nontensor_params) != self.nparams:
            raise ValueError(
                "The total length of tensor and nontensor params "
                "do not match with the expected length: %d instead of %d" %
                (len(tensor_params) + len(nontensor_params), self.nparams))
        if self.alltensors:
            return tensor_params

        params = [None for _ in range(self.nparams)]
        for nidx, p in zip(self.nontensor_idxs, nontensor_params):
            params[nidx] = p
        for idx, p in zip(self.tensor_idxs, tensor_params):
            params[idx] = p
        return params


def tallqr(V, MV=None):
    """QR decomposition for tall and skinny matrix.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.pytorch_utils import tallqr
    >>> V = torch.randn(3, 2)
    >>> Q, R = tallqr(V)
    >>> Q.shape
    torch.Size([3, 2])
    >>> R.shape
    torch.Size([2, 2])
    >>> torch.allclose(Q @ R, V)
    True

    Parameters
    ----------
    V: torch.Tensor
        V is a matrix to be decomposed. (`*BV`, na, nguess)
    MV: torch.Tensor
        (`*BM`, na, nguess) where M is the basis to make Q M-orthogonal
        if MV is None, then MV=V (default=None)

    Returns
    -------
    Q: torch.Tensor
        The Orthogonal Part. Shape: (`*BV`, na, nguess)
    R: torch.Tensor
        The (`*BM`, nguess, nguess) where M is the basis to make Q M-orthogonal

    """
    if MV is None:
        MV = V
    VTV = torch.matmul(V.transpose(-2, -1), MV)  # (*BMV, nguess, nguess)
    R = torch.linalg.cholesky(VTV.transpose(-2, -1).conj()).transpose(
        -2, -1).conj()  # (*BMV, nguess, nguess)
    Rinv = torch.inverse(R)  # (*BMV, nguess, nguess)
    Q = torch.matmul(V, Rinv)
    return Q, R


def to_fortran_order(V):
    """Convert a tensor to Fortran order. (The last two dimensions are made Fortran order.)
    Fortran order/ array is a special case in which all elements of an array are stored in
    column-major order.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.pytorch_utils import to_fortran_order
    >>> V = torch.randn(3, 2)
    >>> V.is_contiguous()
    True
    >>> V = to_fortran_order(V)
    >>> V.is_contiguous()
    False
    >>> V.shape
    torch.Size([3, 2])
    >>> V = torch.randn(3, 2).transpose(-2, -1)
    >>> V.is_contiguous()
    False
    >>> V = to_fortran_order(V)
    >>> V.is_contiguous()
    False
    >>> V.shape
    torch.Size([2, 3])

    Parameters
    ----------
    V: torch.Tensor
        V is a matrix to be converted. (`*BV`, na, nguess)

    Returns
    -------
    outV: torch.Tensor
        (`*BV`, nguess, na)

    """
    if V.is_contiguous():
        # return V.set_(V.storage(), V.storage_offset(), V.size(), tuple(reversed(V.stride())))
        return V.transpose(-2, -1).contiguous().transpose(-2, -1)
    elif V.transpose(-2, -1).is_contiguous():
        return V
    else:
        raise RuntimeError(
            "Only the last two dimensions can be made Fortran order.")


def get_np_dtype(dtype: torch.dtype) -> Any:
    """corresponding numpy dtype from the input pytorch's tensor dtype

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.pytorch_utils import get_np_dtype
    >>> get_np_dtype(torch.float32)
    <class 'numpy.float32'>
    >>> get_np_dtype(torch.float64)
    <class 'numpy.float64'>

    Parameters
    ----------
    dtype: torch.dtype
        pytorch's tensor dtype

    Returns
    -------
    np.dtype
        corresponding numpy dtype

    """
    if dtype == torch.float32:
        return np.float32
    elif dtype == torch.float64:
        return np.float64
    elif dtype == torch.complex64:
        return np.complex64
    elif dtype == torch.complex128:
        return np.complex128
    else:
        raise TypeError("Unknown type: %s" % dtype)


def unsorted_segment_max(data: torch.Tensor, segment_ids: torch.Tensor,
                         num_segments: int) -> torch.Tensor:
    """Computes the maximum along segments of a tensor. Analogous to tf.unsorted_segment_max.

    Parameters
    ----------
    data: torch.Tensor
        A tensor whose segments are to be maximized.
    segment_ids: torch.Tensor
        The segment indices tensor.
    num_segments: int
        The number of segments.

    Returns
    -------
    tensor: torch.Tensor

    Examples
    --------
    >>> segment_ids = torch.Tensor([0, 1, 0]).to(torch.int64)
    >>> data = torch.Tensor([[1, 2, 3, 4], [5, 6, 7, 8], [4, 3, 2, 1]])
    >>> num_segments = 2
    >>> result = unsorted_segment_max(data=data,
    ...                               segment_ids=segment_ids,
    ...                               num_segments=num_segments)
    >>> data.shape[0]
    3
    >>> segment_ids.shape[0]
    3
    >>> len(segment_ids.shape)
    1
    >>> result
    tensor([[4., 3., 3., 4.],
            [5., 6., 7., 8.]])

    """
    if len(segment_ids.shape) != 1:
        raise AssertionError("segment_ids have to be a 1-D tensor")

    if data.shape[0] != segment_ids.shape[0]:
        raise AssertionError(
            "segment_ids should be the same size as dimension 0 of input.")

    # Initialize the tensor to hold the maximum values for each segment
    shape = [num_segments] + list(data.shape[1:])
    tensor = torch.full(shape, float('-inf'), dtype=data.dtype)

    # Create an expanded segment_ids tensor to match data shape
    expanded_segment_ids = segment_ids.unsqueeze(-1).expand(-1, *data.shape[1:])

    # Update the maximum values for each segment
    for i in range(num_segments):
        mask = expanded_segment_ids == i
        tensor[i] = torch.max(data.masked_fill(~mask, float('-inf')), dim=0)[0]

    return tensor


def estimate_ovlp_rcut(precision: float, coeffs: torch.Tensor,
                       alphas: torch.Tensor) -> float:
    """Estimate the rcut for lattice sum to achieve the given precision
    it is estimated based on the overlap integral

    Examples
    --------
    >>> from deepchem.utils import estimate_ovlp_rcut
    >>> precision = 1e-6
    >>> coeffs = torch.tensor([1.0, 2.0, 3.0])
    >>> alphas = torch.tensor([1.0, 2.0, 3.0])
    >>> estimate_ovlp_rcut(precision, coeffs, alphas)
    6.7652716636657715

    Parameters
    ----------
    precision : float
        Precision to be achieved
    coeffs : torch.Tensor
        Coefficients of the basis functions
    alphas : torch.Tensor
        Alpha values of the basis functions

    Returns
    -------
    float
        Estimated rcut
    """
    langmom = 1
    C = (coeffs * coeffs + 1e-200) * (2 * langmom + 1) * alphas / precision
    r0 = torch.tensor(20.0, dtype=coeffs.dtype, device=coeffs.device)
    for i in range(2):
        r0 = torch.sqrt(
            2.0 * torch.log(C *
                            (r0 * r0 * alphas)**(langmom + 1) + 1.) / alphas)
    rcut = float(torch.max(r0).detach())
    return rcut


def get_dtype_memsize(a: torch.Tensor) -> int:
    """Size of each element in the tensor in bytes

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils import get_dtype_memsize
    >>> a = torch.randn(3, 2)
    >>> get_dtype_memsize(a)
    4

    """
    if a.dtype == torch.float64 or a.dtype == torch.int64:
        size = 8
    elif a.dtype == torch.float32 or a.dtype == torch.int32:
        size = 4
    elif a.dtype == torch.bool:
        size = 1
    else:
        raise TypeError("Unknown tensor type: %s" % a.dtype)
    return size



"""
BBBC Dataset loader.

This file contains image loaders for the BBBC dataset collection (https://data.broadinstitute.org/bbbc/image_sets.html).
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union
import zipfile
import numpy as np
import pandas as pd

BBBC1_IMAGE_URL = 'https://data.broadinstitute.org/bbbc/BBBC001/BBBC001_v1_images_tif.zip'
BBBC1_LABEL_URL = 'https://data.broadinstitute.org/bbbc/BBBC001/BBBC001_v1_counts.txt'
BBBC1_TASKS = ["cell-count"]

BBBC2_IMAGE_URL = 'https://data.broadinstitute.org/bbbc/BBBC002/BBBC002_v1_images.zip'
BBBC2_LABEL_URL = 'https://data.broadinstitute.org/bbbc/BBBC002/BBBC002_v1_counts.txt'
BBBC2_TASKS = ["cell-count"]

BBBC3_IMAGE_URL = 'https://data.broadinstitute.org/bbbc/BBBC003/BBBC003_v1_images.zip'
BBBC3_LABEL_URL = 'https://data.broadinstitute.org/bbbc/BBBC003/BBBC003_v1_counts.txt'
BBBC3_FOREGROUND_URL = 'https://data.broadinstitute.org/bbbc/BBBC003/BBBC003_v1_foreground.zip'
BBBC3_TASKS = ["cell-count"]

BBBC4_TASKS = ["cell-count"]

BBBC5_IMAGE_URL = 'https://data.broadinstitute.org/bbbc/BBBC005/BBBC005_v1_images.zip'
BBBC5_FOREGROUND_URL = 'https://data.broadinstitute.org/bbbc/BBBC005/BBBC005_v1_ground_truth.zip'
BBBC5_TASKS = ["cell-count"]


class _BBBC001_Loader(_MolnetLoader):
    """BBBC001 cell count dataset loader"""

    def create_dataset(self) -> Dataset:
        """Creates a dataset from BBBC001 images and cell counts as labels"""
        dataset_file = os.path.join(self.data_dir, "BBBC001_v1_images_tif.zip")
        labels_file = os.path.join(self.data_dir, "BBBC001_v1_counts.txt")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=BBBC1_IMAGE_URL,
                                             dest_dir=self.data_dir)
        if not os.path.exists(labels_file):
            dc.utils.data_utils.download_url(url=BBBC1_LABEL_URL,
                                             dest_dir=self.data_dir)
        labels_table = pd.read_csv(labels_file, delimiter="\t")
        labels = np.mean(
            [labels_table["manual count #1"], labels_table["manual count #2"]],
            axis=0,
            dtype=int)

        loader = dc.data.ImageLoader()
        return loader.create_dataset(inputs=(dataset_file, labels),
                                     in_memory=False)


def load_bbbc001(
    splitter: Union[dc.splits.Splitter, str, None] = 'index',
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load BBBC001 dataset

    This dataset contains 6 images of human HT29 colon cancer cells. The task is
    to learn to predict the cell counts in these images. This dataset is too small
    to serve to train algorithms, but might serve as a good test dataset.
    https://data.broadinstitute.org/bbbc/BBBC001/

    Parameters
    ----------
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    featurizer = dc.feat.UserDefinedFeaturizer([])  # Not actually used
    loader = _BBBC001_Loader(featurizer, splitter, transformers, BBBC1_TASKS,
                             data_dir, save_dir, **kwargs)
    return loader.load_dataset('bbbc001', reload)


class _BBBC002_Loader(_MolnetLoader):
    """BBBC002 cell count dataset loader"""

    def create_dataset(self) -> Dataset:
        """Creates a dataset from BBBC002 images and cell counts as labels"""
        dataset_file = os.path.join(self.data_dir, "BBBC002_v1_images.zip")
        labels_file = os.path.join(self.data_dir, "BBBC002_v1_counts.txt")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=BBBC2_IMAGE_URL,
                                             dest_dir=self.data_dir)
        if not os.path.exists(labels_file):
            dc.utils.data_utils.download_url(url=BBBC2_LABEL_URL,
                                             dest_dir=self.data_dir)

        labels_table = pd.read_csv(labels_file, delimiter="\t")
        labels = np.mean([
            labels_table["human counter 1 (Robert Lindquist)"],
            labels_table["human counter #2 (Joohan Chang)"]
        ],
                         axis=0,
                         dtype=int)

        loader = dc.data.ImageLoader()
        return loader.create_dataset(inputs=(dataset_file, labels),
                                     in_memory=False)


def load_bbbc002(
    splitter: Union[dc.splits.Splitter, str, None] = 'index',
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load BBBC002 dataset

    This dataset contains data corresponding to 5 samples of Drosophilia Kc167
    cells. There are 10 fields of view for each sample, each an image of size
    512x512. Ground truth labels contain cell counts for this dataset. Full
    details about this dataset are present at
    https://data.broadinstitute.org/bbbc/BBBC002/.

    Parameters
    ----------
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    featurizer = dc.feat.UserDefinedFeaturizer([])  # Not actually used
    loader = _BBBC002_Loader(featurizer, splitter, transformers, BBBC2_TASKS,
                             data_dir, save_dir, **kwargs)
    return loader.load_dataset('bbbc002', reload)


class _BBBC003_Segmentation_Loader(_MolnetLoader):
    """BBBC003 segmentation mask dataset loader"""

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "BBBC003_v1_images.zip")
        foreground_file = os.path.join(self.data_dir,
                                       "BBBC003_v1_foreground.zip")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=BBBC3_IMAGE_URL,
                                             dest_dir=self.data_dir)
        if not os.path.exists(foreground_file):
            dc.utils.data_utils.download_url(url=BBBC3_FOREGROUND_URL,
                                             dest_dir=self.data_dir)

        loader = dc.data.ImageLoader(sorting=True)
        return loader.create_dataset(inputs=(dataset_file, foreground_file),
                                     in_memory=False)


class _BBBC003_Loader(_MolnetLoader):
    """BBBC003 cell count dataset loader"""

    def create_dataset(self):
        dataset_file = os.path.join(self.data_dir, "BBBC003_v1_images.zip")
        labels_file = os.path.join(self.data_dir, "BBBC003_v1_counts.txt")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=BBBC3_IMAGE_URL,
                                             dest_dir=self.data_dir)
        if not os.path.exists(labels_file):
            dc.utils.data_utils.download_url(url=BBBC3_LABEL_URL,
                                             dest_dir=self.data_dir)

        labels = pd.read_csv(labels_file, delimiter="\t")
        lbx = labels.sort_values("Image")["manual count #1"].values

        loader = dc.data.ImageLoader(sorting=True)
        return loader.create_dataset(inputs=(dataset_file, lbx),
                                     in_memory=False)


def load_bbbc003(
    load_segmentation_mask: bool = False,
    splitter: Union[dc.splits.Splitter, str, None] = 'index',
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load BBBC003 dataset

    This dataset contains data corresponding to 15 samples of Mouse embryos with DIC.
    Each image is of size 640x480. Ground truth labels contain cell counts and
    segmentation masks for this dataset. Full details about this dataset are present at
    https://data.broadinstitute.org/bbbc/BBBC003/.

    Parameters
    ----------
    load_segmentation_mask: bool
        if True, the dataset will contain segmentation masks as labels. Otherwise,
        the dataset will contain cell counts as labels.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    Examples
    --------
    Importing necessary modules

    >>> import deepchem as dc
    >>> import numpy as np

    We can load the BBBC003 dataset with 2 types of labels: segmentation masks and
    cell counts. We will first load the dataset with cell counts as labels.

    >>> loader = dc.molnet.load_bbbc003(load_segmentation_mask=False)
    >>> tasks, dataset, transformers = loader
    >>> train, val, test = dataset

    We now have a dataset with 15 samples, each with 300 cells. The images are of
    size 640x480. The labels are cell counts. We can verify this as follows:

    >>> train.X.shape
    (12,)
    >>> train.y.shape
    (12,)

    We will now load the dataset with segmentation masks as labels.

    >>> loader = dc.molnet.load_bbbc003(load_segmentation_mask=True)
    >>> tasks, dataset, transformers = loader
    >>> train, val, test = dataset

    We now have a dataset with 15 samples, each with 300 cells. The images are of
    size 640x480. The labels are segmentation masks. We can verify this as follows:

    >>> print(train.X.shape)
    (12,)
    >>> print(train.y.shape)
    (12,)

    Note: The image labelled '7_19_M2E15.tif' is transposed to 480x640 in the source file along with it's
    segementation mask. To match it with the other images, we need to transpose it back to 640x480.

    This image is found at index 6 in the train dataset (Assuming no shuffling has taken place).

    First, we load the dataset as usual and split it into X, y, w and ids. Here, X is the list
    of input images, y is the list of labels, w is the list of weights and ids is the list of
    IDs for each sample.

    >>> train_x, train_y, train_w, train_ids = train.X, train.y, train.w, train.ids

    We can now transpose the image at index 6 in the input data (train_x):
    >>> train_x[6] = train_x[6].T

    We can now verify that the image is of size 640x480:
    >>> print(train_x[6].shape)
    (640, 480)

    This is also seen in the segmentation mask with the same filename and index, in which
    case, we transpose the label (train_y) instead of the input data:

    >>> train_y[6] = train_y[6].T

    We can now verify that the image is of size 640x480:
    >>> train_y[6].shape
    (640, 480)
    """
    featurizer = dc.feat.UserDefinedFeaturizer([])  # Not actually used
    loader: _MolnetLoader
    if load_segmentation_mask:
        loader = _BBBC003_Segmentation_Loader(featurizer, splitter,
                                              transformers, BBBC3_TASKS,
                                              data_dir, save_dir, **kwargs)
    else:
        loader = _BBBC003_Loader(featurizer, splitter, transformers,
                                 BBBC3_TASKS, data_dir, save_dir, **kwargs)

    return loader.load_dataset('bbbc003', reload)


class _BBBC004_Segmentation_Loader(_MolnetLoader):
    """BBBC004 segmentation mask dataset loader"""

    def __init__(self, overlap_probability, featurizer, splitter, transformers,
                 BBBC4_TASKS, data_dir, save_dir, **kwargs):
        overlap_dict = {
            0.0: "000",
            0.15: "015",
            0.3: "030",
            0.45: "045",
            0.6: "060"
        }
        if overlap_probability not in overlap_dict.keys():
            raise ValueError(
                f"Overlap_probability must be one of {overlap_dict.keys()}, got {overlap_probability}"
            )
        else:
            self.overlap_probability = overlap_dict[overlap_probability]
            self.BBBC4_IMAGE_URL = f"https://data.broadinstitute.org/bbbc/BBBC004/BBBC004_v1_{self.overlap_probability}_images.zip"
            self.BBBC4_FOREGROUND_URL = f"https://data.broadinstitute.org/bbbc/BBBC004/BBBC004_v1_{self.overlap_probability}_foreground.zip"

        super(_BBBC004_Segmentation_Loader,
              self).__init__(featurizer, splitter, transformers, BBBC4_TASKS,
                             data_dir, save_dir, **kwargs)

    def create_dataset(self) -> Dataset:
        """Creates a dataset from BBBC004 images and segmentation masks as labels"""

        dataset_file = os.path.join(
            self.data_dir, f"BBBC004_v1_{self.overlap_probability}_images.zip")
        foreground_file = os.path.join(
            self.data_dir,
            f"BBBC004_v1_{self.overlap_probability}_foreground.zip")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=self.BBBC4_IMAGE_URL,
                                             dest_dir=self.data_dir)
        if not os.path.exists(foreground_file):
            dc.utils.data_utils.download_url(url=self.BBBC4_FOREGROUND_URL,
                                             dest_dir=self.data_dir)

        loader = dc.data.ImageLoader(sorting=False)
        return loader.create_dataset(inputs=(dataset_file, foreground_file),
                                     in_memory=False)


class _BBBC004_Loader(_MolnetLoader):
    """BBBC004 cell count dataset loader"""

    def __init__(self, overlap_probability, featurizer, splitter, transformers,
                 BBBC4_TASKS, data_dir, save_dir, **kwargs):
        overlap_dict = {
            0.0: "000",
            0.15: "015",
            0.3: "030",
            0.45: "045",
            0.6: "060"
        }
        if overlap_probability not in overlap_dict.keys():
            raise ValueError(
                f"Overlap_probability must be one of {overlap_dict.keys()}, got {overlap_probability}"
            )
        else:
            self.overlap_probability = overlap_dict[overlap_probability]
            self.BBBC4_IMAGE_URL = f"https://data.broadinstitute.org/bbbc/BBBC004/BBBC004_v1_{self.overlap_probability}_images.zip"
            self.BBBC4_FOREGROUND_URL = f"https://data.broadinstitute.org/bbbc/BBBC004/BBBC004_v1_{self.overlap_probability}_foreground.zip"

        super(_BBBC004_Loader,
              self).__init__(featurizer, splitter, transformers, BBBC4_TASKS,
                             data_dir, save_dir, **kwargs)

    def create_dataset(self) -> Dataset:
        """Creates a dataset from BBBC004 images and cell counts as labels"""

        dataset_file = os.path.join(
            self.data_dir, f"BBBC004_v1_{self.overlap_probability}_images.zip")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=self.BBBC4_IMAGE_URL,
                                             dest_dir=self.data_dir)
        labels = np.full(20, 300, dtype=int)

        loader = dc.data.ImageLoader(sorting=False)
        return loader.create_dataset(inputs=(dataset_file, labels),
                                     in_memory=False)


def load_bbbc004(
    overlap_probability: float = 0.0,
    load_segmentation_mask: bool = False,
    splitter: Union[dc.splits.Splitter, str, None] = 'index',
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load BBBC004 dataset

    This dataset contains data corresponding to 20 samples of synthetically generated
    fluorescent cell population images. There are 300 cells in each sample, each an image
    of size 950x950. Ground truth labels contain cell counts and segmentation masks for
    this dataset. Full details about this dataset are present at
    https://data.broadinstitute.org/bbbc/BBBC004/.

    Parameters
    ----------
    overlap_probability: float from list {0.0, 0.15, 0.3, 0.45, 0.6}
        the overlap probability of the synthetic cells in the images
    load_segmentation_mask: bool
        if True, the dataset will contain segmentation masks as labels. Otherwise,
        the dataset will contain cell counts as labels.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    Examples
    --------
    Importing necessary modules

    >>> import deepchem as dc
    >>> import numpy as np

    We can load the BBBC004 dataset with 2 types of labels: segmentation masks and
    cell counts. We will first load the dataset with cell counts as labels.

    >>> loader = dc.molnet.load_bbbc004(overlap_probability=0.0, load_segmentation_mask=False)
    >>> tasks, dataset, transformers = loader
    >>> train, val, test = dataset

    We now have a dataset with 20 samples, each with 300 cells. The images are of
    size 950x950. The labels are cell counts. We can verify this as follows:

    >>> train.X.shape
    (16, 950, 950)
    >>> train.y.shape
    (16,)

    We will now load the dataset with segmentation masks as labels.

    >>> loader = dc.molnet.load_bbbc004(overlap_probability=0.0, load_segmentation_mask=True)
    >>> tasks, dataset, transformers = loader
    >>> train, val, test = dataset

    We now have a dataset with 20 samples, each with 300 cells. The images are of
    size 950x950. The labels are segmentation masks. We can verify this as follows:

    >>> train.X.shape
    (16, 950, 950)
    >>> train.y.shape
    (16, 950, 950, 3)
    """
    featurizer = dc.feat.UserDefinedFeaturizer([])  # Not actually used
    loader: _MolnetLoader
    if load_segmentation_mask:
        loader = _BBBC004_Segmentation_Loader(overlap_probability, featurizer,
                                              splitter, transformers,
                                              BBBC4_TASKS, data_dir, save_dir,
                                              **kwargs)
    else:
        loader = _BBBC004_Loader(overlap_probability, featurizer, splitter,
                                 transformers, BBBC4_TASKS, data_dir, save_dir,
                                 **kwargs)

    return loader.load_dataset('bbbc004', reload)


class _BBBC005_Loader(_MolnetLoader):
    """BBBC005 cell count dataset loader"""

    def create_dataset(self):
        dataset_file = os.path.join(self.data_dir, "BBBC005_v1_images.zip")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=BBBC5_IMAGE_URL,
                                             dest_dir=self.data_dir)

        labels = []

        # Read the zip file
        with zipfile.ZipFile(dataset_file, 'r') as zip_ref:
            file_list = zip_ref.namelist()

        file_list = file_list[1:]

        # Get the labels from filenames
        for filename in file_list:
            if filename.split('/')[-1].split('.')[-1] == 'TIF':
                labels.append(int(filename.split('/')[-1].split('_')[2][1:]))

        lbx = np.array(labels, dtype=np.int32)

        loader = dc.data.ImageLoader(sorting=False)
        return loader.create_dataset(inputs=(dataset_file, lbx),
                                     in_memory=False)


def load_bbbc005(
    splitter: Union[dc.splits.Splitter, str, None] = 'index',
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load BBBC005 dataset

    This dataset contains data corresponding to 19,200 samples of synthetically generated
    fluorescent cell population images. These images were simulated for a given cell count
    with a clustering probablity of 25% and a CCD noise variance of 0.0001. Focus blur
    was simulated by applying varying Guassian filters to the images. Each image is of
    size 520x696. Ground truth labels contain cell counts for this dataset. Full details
    about this dataset are present at
    https://data.broadinstitute.org/bbbc/BBBC005/.

    Parameters
    ----------
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    Examples
    --------
    Importing necessary modules

    >> import deepchem as dc
    >> import numpy as np

    We will now load the BBBC005 dataset with cell counts as labels.

    >> loader = dc.molnet.load_bbbc005()
    >> tasks, dataset, transformers = loader
    >> train, val, test = dataset

    We now have a dataset with a total of 19,200 samples with cell counts in
    the range of 1-100. The images are of size 520x696. The labels are cell
    counts. We have a train-val-test split of 80:10:10. We can verify this as follows:

    >> train.X.shape
    (15360, 520, 696)
    >> train.y.shape
    (15360,)
    """
    featurizer = dc.feat.UserDefinedFeaturizer([])  # Not actually used
    loader: _MolnetLoader
    loader = _BBBC005_Loader(featurizer, splitter, transformers, BBBC5_TASKS,
                             data_dir, save_dir, **kwargs)

    return loader.load_dataset('bbbc005', reload)



"""
Implementation of MEGNet class
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

from deepchem.models.losses import Loss, L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.torch_models.layers import GraphNetwork as GN
from deepchem.models.torch_models import TorchModel


class MEGNet(nn.Module):
    """MatErials Graph Network

    A model for predicting crystal and molecular properties using GraphNetworks.

    Example
    -------
    >>> import numpy as np
    >>> from torch_geometric.data import Batch
    >>> from deepchem.feat import GraphData
    >>> n_nodes, n_node_features = 5, 10
    >>> n_edges, n_edge_attrs = 5, 2
    >>> n_global_features = 4
    >>> node_features = np.random.randn(n_nodes, n_node_features)
    >>> edge_attrs = np.random.randn(n_edges, n_edge_attrs)
    >>> edge_index = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]], dtype=np.int64)
    >>> global_features = np.random.randn(1, n_global_features)
    >>> graph = GraphData(node_features, edge_index, edge_attrs, global_features=global_features)
    >>> batch = Batch()
    >>> batch = batch.from_data_list([graph.to_pyg_graph()])
    >>> model = MEGNet(n_node_features=n_node_features, n_edge_features=n_edge_attrs, n_global_features=n_global_features)
    >>> pred = model(batch)

    Note
    ----
    This class requires torch-geometric to be installed.
    """

    def __init__(self,
                 n_node_features: int = 32,
                 n_edge_features: int = 32,
                 n_global_features: int = 32,
                 n_blocks: int = 1,
                 is_undirected: bool = True,
                 residual_connection: bool = True,
                 mode: str = 'regression',
                 n_classes: int = 2,
                 n_tasks: int = 1):
        """

        Parameters
        ----------
        n_node_features: int
            Number of features in a node
        n_edge_features: int
            Number of features in a edge
        n_global_features: int
            Number of global features
        n_blocks: int
            Number of GraphNetworks block to use in update
        is_undirected: bool, optional (default True)
            True when the graph is undirected graph , otherwise False
        residual_connection: bool, optional (default True)
            If True, the layer uses a residual connection during training
        n_tasks: int, default 1
            The number of tasks
        mode: str, default 'regression'
            The model type - classification or regression
        n_classes: int, default 2
            The number of classes to predict (used only in classification mode).
        """
        super(MEGNet, self).__init__()
        try:
            from torch_geometric.nn import Set2Set
        except ModuleNotFoundError:
            raise ImportError(
                "MEGNet model requires torch_geometric to be installed")

        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")
        self.n_node_features = n_node_features
        self.n_edge_features = n_edge_features
        self.n_global_features = n_global_features
        self.megnet_blocks = nn.ModuleList()
        self.n_blocks = n_blocks
        for i in range(n_blocks):
            self.megnet_blocks.append(
                GN(n_node_features=n_node_features,
                   n_edge_features=n_edge_features,
                   n_global_features=n_global_features,
                   is_undirected=is_undirected,
                   residual_connection=residual_connection))
        self.n_tasks = n_tasks
        self.mode = mode
        self.n_classes = n_classes

        self.set2set_nodes = Set2Set(in_channels=n_node_features,
                                     processing_steps=3,
                                     num_layers=1)
        self.set2set_edges = Set2Set(in_channels=n_edge_features,
                                     processing_steps=3,
                                     num_layers=1)

        self.dense = nn.Sequential(
            nn.Linear(in_features=2 * n_node_features + 2 * n_edge_features +
                      n_global_features,
                      out_features=32),
            nn.Linear(in_features=32, out_features=16))

        if self.mode == 'regression':
            self.out = nn.Linear(in_features=16, out_features=n_tasks)
        elif self.mode == 'classification':
            self.out = nn.Linear(in_features=16,
                                 out_features=n_tasks * n_classes)

    def forward(self, pyg_batch):
        """
        Parameters
        ----------
        pyg_batch: torch_geometric.data.Batch
            A pytorch-geometric batch of graphs where node attributes are stores
            as pyg_batch['x'], edge_index in pyg_batch['edge_index'], edge features
            in pyg_batch['edge_attr'], global features in pyg_batch['global_features']

        Returns
        -------
        torch.Tensor: Predictions for the graph
        """
        node_features = pyg_batch['x']
        edge_index, edge_features = pyg_batch['edge_index'], pyg_batch[
            'edge_attr']
        global_features = pyg_batch['global_features']
        batch = pyg_batch['batch']

        for i in range(self.n_blocks):
            node_features, edge_features, global_features = self.megnet_blocks[
                i](node_features, edge_index, edge_features, global_features,
                   batch)

        node_features = self.set2set_nodes(node_features, batch)
        edge_features = self.set2set_edges(edge_features, batch[edge_index[0]])
        out = torch.cat([node_features, edge_features, global_features], axis=1)
        out = self.out(self.dense(out))

        if self.mode == 'classification':
            if self.n_tasks == 1:
                logits = out.view(-1, self.n_classes)
                softmax_dim = 1
            else:
                logits = out.view(-1, self.n_tasks, self.n_classes)
                softmax_dim = 2
            proba = F.softmax(logits, dim=softmax_dim)
            return proba, logits
        elif self.mode == 'regression':
            return out


class MEGNetModel(TorchModel):
    """MatErials Graph Network for Molecules and Crystals

    MatErials Graph Network [1]_ are Graph Networks [2]_ which are used for property prediction
    in molecules and crystals. The model implements multiple layers of Graph Network as
    MEGNetBlocks and then combines the node properties and edge properties of all nodes
    and edges via a Set2Set layer. The combines information is used with the global
    features of the material/molecule for property prediction tasks.

    Example
    -------
    >>> import deepchem as dc
    >>> from deepchem.models import MEGNetModel
    >>> from deepchem.utils.fake_data_generator import FakeGraphGenerator as FGG
    >>> graphs = FGG(global_features=4, num_classes=10).sample(n_graphs=20)
    >>> model = dc.models.MEGNetModel(n_node_features=5, n_edge_features=3, n_global_features=4, n_blocks=3, is_undirected=True, residual_connection=True, mode='classification', n_classes=10, batch_size=16)
    >>> training_loss = model.fit(graphs)

    References
    ----------
    .. [1] Chen, Chi, et al. "Graph networks as a universal machine learning framework for molecules and crystals." Chemistry of Materials 31.9 (2019): 3564-3572.
    .. [2] Battaglia, Peter W., et al. "Relational inductive biases, deep learning, and graph networks." arXiv preprint arXiv:1806.01261 (2018).
    Note
    ----
    The model requires PyTorch-Geometric to be installed.
    """

    def __init__(self,
                 n_node_features: int = 32,
                 n_edge_features: int = 32,
                 n_global_features: int = 32,
                 n_blocks: int = 1,
                 is_undirected: bool = True,
                 residual_connection: bool = True,
                 mode: str = 'regression',
                 n_classes: int = 2,
                 n_tasks: int = 1,
                 **kwargs):
        """
        Parameters
        ----------
        n_node_features: int
            Number of features in a node
        n_edge_features: int
            Number of features in a edge
        n_global_features: int
            Number of global features
        n_blocks: int
            Number of GraphNetworks block to use in update
        is_undirected: bool, optional (default True)
            True when the model is used on undirected graphs otherwise false
        residual_connection: bool, optional (default True)
            If True, the layer uses a residual connection during training
        n_tasks: int, default 1
            The number of tasks
        mode: str, default 'regression'
            The model type - classification or regression
        n_classes: int, default 2
            The number of classes to predict (used only in classification mode).
        kwargs: Dict
            kwargs supported by TorchModel
        """
        model = MEGNet(n_node_features=n_node_features,
                       n_edge_features=n_edge_features,
                       n_global_features=n_global_features,
                       n_blocks=n_blocks,
                       is_undirected=is_undirected,
                       residual_connection=residual_connection,
                       mode=mode,
                       n_classes=n_classes,
                       n_tasks=n_tasks)
        if mode == 'regression':
            loss: Loss = L2Loss()
            output_types = ['prediction']
        elif mode == 'classification':
            loss = SparseSoftmaxCrossEntropy()
            output_types = ['prediction', 'loss']
        super(MEGNetModel, self).__init__(model,
                                          loss=loss,
                                          output_types=output_types,
                                          **kwargs)

    def _prepare_batch(self, batch):
        """Creates batch data for MEGNet model

        Note
        ----
        Ideally, we should only override default_generator method. But the problem
        here is that we _prepare_batch of TorchModel only supports non-graph
        data types. Hence, we are overriding it here. This should be fixed
        some time in the future.
        """
        try:
            from torch_geometric.data import Batch
        except ModuleNotFoundError:
            raise ImportError("This module requires PyTorch Geometric")

        # We convert deepchem.feat.GraphData to a PyG graph and then
        # batch it.
        graphs, labels, weights = batch
        # The default_generator method returns an array of dc.feat.GraphData objects
        # nested inside a list. To access the nested array of graphs, we are
        # indexing by 0 here.
        graph_list = [graph.to_pyg_graph() for graph in graphs[0]]
        pyg_batch = Batch()
        pyg_batch = pyg_batch.from_data_list(graph_list)

        _, labels, weights = super(MEGNetModel, self)._prepare_batch(
            ([], labels, weights))

        return pyg_batch, labels, weights



import logging
import os
import torch
import warnings
from typing import List
from deepchem.utils.dft_utils import CGTOBasis
try:
    import basis_set_exchange as bse
except Exception as e:
    warnings.warn(f'basis-set-exchange Not Found: {e}')

logger = logging.getLogger(__name__)
_dtype = torch.double
_device = torch.device("cpu")


def loadbasis(cmd: str, dtype: torch.dtype = _dtype,
              device: torch.device = _device, requires_grad: bool = False) -> \
        List[CGTOBasis]:
    """
    Load basis from a file and return the list of CGTOBasis.

    Examples
    --------
    >>> H = loadbasis("1:3-21G")
    >>> H[0].alphas.shape
    torch.Size([2])

    Parameters
    ----------
    cmd: str
        This can be a file path where the basis is stored or a
        string in format ``"atomz:basis"``, e.g. ``"1:6-311++G**"``.
    dtype: torch.dtype
        Tensor data type for ``alphas`` and ``coeffs`` of the GTO basis
    device: torch.device
        Tensor device for ``alphas`` and ``coeffs``
    requires_grad: bool
        If ``True``, the ``alphas`` and ``coeffs`` tensors become differentiable

    Returns
    -------
    list of CGTOBasis
        List of GTO basis loaded from the given file

    """
    res = []
    if not os.path.exists(cmd):
        file = _get_basis_file(cmd)
    else:
        file = cmd

    # read the content
    with open(file, "r") as f:
        lines = f.read().split("\n")

    # skip the header
    while True:
        line = lines.pop(0)
        if line == "":
            continue
        if line.startswith("!"):
            continue
        break

    # now it is at the orbital description
    while len(lines) > 0:
        line = lines.pop(0)
        if line.startswith("**"):
            break
        desc = line.split()
        nlines = int(desc[1])
        if nlines == 0:
            raise RuntimeError("Zero line on basis %s" % file)

        # read the exponents and the coefficients
        alphas = []
        coeffsT = []
        for i in range(nlines):
            alphacoeff = [_read_float(f) for f in lines.pop(0).split()]
            alphas.append(alphacoeff[0])
            coeffsT.append(alphacoeff[1:])
        # coeffsT: list with shape (nbasis, ncontr)
        # coeffs: list with shape (ncontr, nbasis)
        coeffs = list(zip(*coeffsT))
        ncoeffs = len(coeffs)
        angmoms = _expand_angmoms(desc[0], ncoeffs)

        # convert to tensor
        alpha = torch.tensor(alphas,
                             dtype=dtype,
                             device=device,
                             requires_grad=requires_grad)
        for i in range(ncoeffs):
            coeff = torch.tensor(coeffs[i],
                                 dtype=dtype,
                                 device=device,
                                 requires_grad=requires_grad)
            basis = CGTOBasis(angmom=angmoms[i], alphas=alpha, coeffs=coeff)
            basis.wfnormalize_()
            res.append(basis)
    return res


def _read_float(s: str) -> float:
    """Read a float from a string, replacing D with E.

    Examples
    --------
    >>> _read_float("1.0D+00")
    1.0

    Parameters
    ----------
    s: str
        String to be converted to float

    Returns
    -------
    float
        Float number

    """
    s = s.replace("D", "E")
    return float(s)


def _get_basis_file(cmd: str) -> str:
    """parse the string command, check if the basis has already been downloaded
    (download if not), and return the file name

    Examples
    --------
    >>> path = _get_basis_file("1:6-311++G**")

    Parameters
    ----------
    cmd: str
        Command string in the format "atomz:basis"

    Returns
    -------
    str
        File path to the basis set

    """

    # parse to get the atomz and the basisname
    atomz_str, raw_basisname = cmd.split(":")
    raw_basisname = raw_basisname.strip()
    atomz = int(atomz_str)

    # get the path to the database
    basisname = _normalize_basisname(raw_basisname)
    thisdir = os.path.dirname(os.path.realpath(__file__))
    fname = "%02d.gaussian94" % atomz
    fdir = os.path.join(thisdir, ".database", basisname)
    fpath = os.path.join(fdir, fname)

    # if the file does not exist, download it
    if not os.path.exists(fpath):
        logger.info(
            "The %s basis for atomz %d does not exist, but we will download it"
            % (raw_basisname, atomz))
        if not os.path.exists(fdir):
            os.makedirs(fdir)
        _download_basis(fpath, atomz, raw_basisname)

    return fpath


def _normalize_basisname(basisname: str) -> str:
    """Normalize the basis name to a file name

    Examples
    --------
    >>> _normalize_basisname("6-311++G**")
    '6-311ppgss'

    Parameters
    ----------
    basisname: str
        Basis name to be normalized

    Returns
    -------
    str
        Normalized basis name

    """
    b = basisname.lower()
    b = b.replace("+", "p")
    b = b.replace("*", "s")
    b = b.replace("(", "_")
    b = b.replace(")", "_")
    b = b.replace(",", "_")
    return b


def _download_basis(fname: str, atomz: int, basisname: str) -> str:
    """Download the basis set from the basis set exchange and save it to a file

    Examples
    --------
    >>> import pathlib
    >>> import os
    >>> path = pathlib.Path(__file__).parent.resolve()
    >>> dir = str(path) + "/.database/6-311ppgss"
    >>> if not os.path.exists(dir):
    ...     os.makedirs(dir)
    >>> path = _download_basis(str(path) + "/01.gaussian94", 1, "6-311++G**")

    Parameters
    ----------
    fname: str
        File name to save the basis set
    atomz: int
        Atomic number of the atom
    basisname: str
        Basis name to be downloaded

    """
    s = bse.get_basis(basisname, elements=[atomz], fmt="gaussian94")
    with open(fname, "w") as f:
        f.write(s)
    return fname


def _expand_angmoms(s: str, n: int) -> List[int]:
    """convert the angular momentum characters into angmom and returns a list
    of n integer containing the angular momentums

    Examples
    --------
    >>> _expand_angmoms("SP", 2)
    [0, 1]

    Parameters
    ----------
    s: str
        Angular momentum characters
    n: int
        Number of coefficients

    Returns
    -------
    List[int]
        List of angular momentums

    """
    if len(s) == n:
        pass
    elif n % len(s) == 0:
        s = s * (n // len(s))
    else:
        raise RuntimeError(
            "Do not know how to read orbital %s with %d coefficient columns" %
            (s, n))
    s = s.lower()
    spdfmap = {
        "s": 0,
        "p": 1,
        "d": 2,
        "f": 3,
        "g": 4,
        "h": 5,
        "i": 6,
    }
    angmoms = [spdfmap[c] for c in s]
    return angmoms



"""
Density Functional Theory utilities.
"""
# flake8: noqa
import logging

logger_ = logging.getLogger(__name__)

from deepchem.utils.dft_utils.config import config

try:
    from deepchem.utils.dft_utils.hamilton.intor.lattice import Lattice

    from deepchem.utils.dft_utils.data.datastruct import ZType
    from deepchem.utils.dft_utils.data.datastruct import AtomPosType
    from deepchem.utils.dft_utils.data.datastruct import AtomZsType
    from deepchem.utils.dft_utils.data.datastruct import SpinParam
    from deepchem.utils.dft_utils.data.datastruct import ValGrad
    from deepchem.utils.dft_utils.data.datastruct import CGTOBasis
    from deepchem.utils.dft_utils.data.datastruct import AtomCGTOBasis
    from deepchem.utils.dft_utils.data.datastruct import BasisInpType
    from deepchem.utils.dft_utils.data.datastruct import DensityFitInfo
    from deepchem.utils.dft_utils.data.datastruct import is_z_float

    from deepchem.utils.dft_utils.hamilton.orbconverter import OrbitalOrthogonalizer
    from deepchem.utils.dft_utils.hamilton.orbparams import BaseOrbParams
    from deepchem.utils.dft_utils.hamilton.orbparams import QROrbParams
    from deepchem.utils.dft_utils.hamilton.orbparams import MatExpOrbParams
    from deepchem.utils.dft_utils.hamilton.intor.lcintwrap import LibcintWrapper
    from deepchem.utils.dft_utils.hamilton.intor.lcintwrap import SubsetLibcintWrapper
    from deepchem.utils.dft_utils.hamilton.intor.molintor import int1e
    from deepchem.utils.dft_utils.hamilton.intor.molintor import int2c2e
    from deepchem.utils.dft_utils.hamilton.intor.molintor import int3c2e
    from deepchem.utils.dft_utils.hamilton.intor.molintor import int2e
    from deepchem.utils.dft_utils.hamilton.intor.molintor import overlap
    from deepchem.utils.dft_utils.hamilton.intor.molintor import kinetic
    from deepchem.utils.dft_utils.hamilton.intor.molintor import nuclattr
    from deepchem.utils.dft_utils.hamilton.intor.molintor import elrep
    from deepchem.utils.dft_utils.hamilton.intor.molintor import coul2c
    from deepchem.utils.dft_utils.hamilton.intor.molintor import coul3c
    from deepchem.utils.dft_utils.hamilton.intor.pbcintor import PBCIntOption
    from deepchem.utils.dft_utils.hamilton.intor.pbcintor import get_default_options
    from deepchem.utils.dft_utils.hamilton.intor.pbcintor import get_default_kpts
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import evl
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import pbc_evl
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import eval_gto
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import eval_gradgto
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import eval_laplgto
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import pbc_eval_gto
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import pbc_eval_gradgto
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import pbc_eval_laplgto
    from deepchem.utils.dft_utils.hamilton.intor.gtoeval import gto_evaluator
    from deepchem.utils.dft_utils.df.base_df import BaseDF
    from deepchem.utils.dft_utils.df.dfmol import DFMol

    from deepchem.utils.dft_utils.api.parser import parse_moldesc
    from deepchem.utils.dft_utils.api.loadbasis import loadbasis

    from deepchem.utils.dft_utils.grid.base_grid import BaseGrid
    from deepchem.utils.dft_utils.grid.radial_grid import RadialGrid
    from deepchem.utils.dft_utils.grid.truncation_rules import BaseTruncationRules
    from deepchem.utils.dft_utils.grid.truncation_rules import NoTrunc
    from deepchem.utils.dft_utils.grid.truncation_rules import DasguptaTrunc
    from deepchem.utils.dft_utils.grid.truncation_rules import NWChemTrunc
    from deepchem.utils.dft_utils.grid.lebedev_grid import LebedevLoader
    from deepchem.utils.dft_utils.grid.lebedev_grid import LebedevGrid
    from deepchem.utils.dft_utils.grid.lebedev_grid import TruncatedLebedevGrid
    from deepchem.utils.dft_utils.grid.multiatoms_grid import BeckeGrid
    from deepchem.utils.dft_utils.grid.multiatoms_grid import PBCBeckeGrid
    from deepchem.utils.dft_utils.grid.radial_grid import get_xw_integration
    from deepchem.utils.dft_utils.grid.radial_grid import SlicedRadialGrid
    from deepchem.utils.dft_utils.grid.radial_grid import BaseGridTransform
    from deepchem.utils.dft_utils.grid.radial_grid import DE2Transformation
    from deepchem.utils.dft_utils.grid.radial_grid import LogM3Transformation
    from deepchem.utils.dft_utils.grid.radial_grid import TreutlerM4Transformation
    from deepchem.utils.dft_utils.grid.radial_grid import get_grid_transform
    from deepchem.utils.dft_utils.grid.factory import get_grid
    from deepchem.utils.dft_utils.grid.factory import get_predefined_grid

    from deepchem.utils.dft_utils.xc.base_xc import BaseXC
    from deepchem.utils.dft_utils.xc.base_xc import AddBaseXC
    from deepchem.utils.dft_utils.xc.base_xc import MulBaseXC
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcLDALibXCPol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcLDALibXCUnpol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcGGALibXCUnpol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcGGALibXCPol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcMGGALibXCUnpol
    from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcMGGALibXCPol

    from deepchem.utils.dft_utils.xc.libxc import LibXCLDA
    from deepchem.utils.dft_utils.xc.libxc import LibXCGGA
    from deepchem.utils.dft_utils.xc.libxc import LibXCMGGA

    from deepchem.utils.dft_utils.api.getxc import get_libxc
    from deepchem.utils.dft_utils.api.getxc import get_xc

    from deepchem.utils.dft_utils.hamilton.base_hamilton import BaseHamilton
    from deepchem.utils.dft_utils.hamilton.hcgto import HamiltonCGTO

    from deepchem.utils.dft_utils.system.base_system import BaseSystem
    from deepchem.utils.dft_utils.system.mol import Mol

    from deepchem.utils.dft_utils.qccalc.base_qccalc import BaseQCCalc
    from deepchem.utils.dft_utils.qccalc.scf_qccalc import SCF_QCCalc
    from deepchem.utils.dft_utils.qccalc.scf_qccalc import BaseSCFEngine
    from deepchem.utils.dft_utils.qccalc.hf import HF
    from deepchem.utils.dft_utils.qccalc.hf import HFEngine
    from deepchem.utils.dft_utils.qccalc.ks import KS
    from deepchem.utils.dft_utils.qccalc.ks import KSEngine
except ModuleNotFoundError as e:
    logger_.warning(
        f'Skipped loading some Pytorch utilities, missing a dependency. {e}')
